//go:build !test
// +build !test

package database

import (
	log "asapo_common/logger"
	"asapo_common/utils"
	"context"
	"encoding/json"
	"errors"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"math"
	"strings"
	"sync"
	"time"
)

type AckRecord struct {
	Id      int    `bson:"message_id"`
	Streams string `bson:"stream" json:"stream"`
	GroupId string `bson:"group_id" json:"group_id"`
}

type MessageRecord struct {
	ID             int                    `json:"message_id"`
	TimeID         int                    `json:"time_id"`
	Timestamp      int                    `bson:"timestamp" json:"timestamp"`
	Name           string                 `json:"name"`
	Meta           map[string]interface{} `json:"meta"`
	NextStream     string
	FinishedStream bool
}

type InProcessingRecord struct {
	Id                int    `bson:"message_id" json:"_id"`
	Streams           string `bson:"stream" json:"stream"`
	GroupId           string `bson:"group_id" json:"group_id"`
	MaxResendAttempts int    `bson:"maxResendAttempts" json:"maxResendAttempts"`
	ResendAttempts    int    `bson:"resendAttempts" json:"resendAttempts"`
	DelayMs           int64  `bson:"delayMs" json:"delayMs"`
}

type NegAckParamsRecord struct {
	ID                int   `bson:"_id" json:"_id"`
	MaxResendAttempts int   `bson:"maxResendAttempts" json:"maxResendAttempts"`
	ResendAttempts    int   `bson:"resendAttempts" json:"resendAttempts"`
	DelayMs           int64 `bson:"delayMs" json:"delayMs"`
}

type Nacks struct {
	Unacknowledged []int `json:"unacknowledged"`
}

type Sources struct {
	Datasources []string `json:"sources"`
}

type LastAck struct {
	ID int `bson:"message_id" json:"lastAckId"`
}

type LocationPointer struct {
	GroupID string `bson:"_id"` // Contain group_id and stream
	Value   int    `bson:"current_pointer"`
}

type PersistedStreamsList struct {
	ID      int      `json:"_id"`
	Streams []string `json:"persisted_streams"`
}

const data_collection_name_prefix = "data_"
const acks_collection_name_prefix = "acks_"
const inprocess_collection_name_prefix = "inprocess_"
const meta_collection_name = "meta"
const streams_info = "streams"
const pointer_collection_name = "current_location"
const pointer_field_name = "current_pointer"
const last_message_collection_name = "last_messages"
const last_message_field_name = "last_message"
const auto_id_counter_name = "auto_id_counters"

const no_session_msg = "database client not created"
const already_connected_msg = "already connected"
const failed_extract_extra_params = "failed to extract extra parameters"
const too_many_persisted_streams = "too many persisted streams"

const finish_stream_keyword = "asapo_finish_stream"
const no_next_stream_keyword = "asapo_no_next"
const Stream_filter_all = "all"
const Stream_filter_finished = "finished"
const Stream_filter_unfinished = "unfinished"

const (
	field_op_inc int = iota
	field_op_set
)

type fieldChangeRequest struct {
	collectionName string
	fieldName      string
	op             int
	max_ind        int
	val            int
}

var dbSessionLock sync.Mutex
var dbClientLock sync.RWMutex

type SizeRecord struct {
	Size int `bson:"size" json:"size"`
}

type Mongodb struct {
	client                *mongo.Client
	timeout               time.Duration
	settings              DBSettings
	lastReadFromInprocess map[string]int64
}

func (db *Mongodb) SetSettings(settings DBSettings) {
	db.settings = settings
}

func (db *Mongodb) Ping() (err error) {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}
	ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
	defer cancel()
	return db.client.Ping(ctx, nil)
}

func (db *Mongodb) Connect(address string) (err error) {
	dbClientLock.Lock()
	defer dbClientLock.Unlock()

	if db.client != nil {
		return &DBError{utils.StatusServiceUnavailable, already_connected_msg}
	}
	db.client, err = mongo.NewClient(options.Client().SetConnectTimeout(20 * time.Second).ApplyURI("mongodb://" + address))
	if err != nil {
		return err
	}
	ctx, cancel := context.WithTimeout(context.Background(), 20*time.Second)
	defer cancel()
	err = db.client.Connect(ctx)
	if err != nil {
		db.client = nil
		return err
	}

	if db.lastReadFromInprocess == nil {
		db.lastReadFromInprocess = make(map[string]int64, 100)
	}

	return db.Ping()
}

func (db *Mongodb) Close() {
	dbClientLock.Lock()
	defer dbClientLock.Unlock()
	if db.client != nil {
		ctx, cancel := context.WithTimeout(context.Background(), 2*time.Second)
		defer cancel()
		db.client.Disconnect(ctx)
		db.client = nil
	}
}

func (db *Mongodb) DropDatabase(dbname string) (err error) {
	return db.dropDatabase(dbname)
}

// This function is used in tests only
func (db *Mongodb) dropDatabase(dbname string) (err error) {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}
	return db.client.Database(dbname).Drop(context.TODO())
}

// This function is used in tests only
// One have to take care to create index if needed.
func (db *Mongodb) insertRecord(dbname string, s interface{}) error {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}

	c := db.client.Database(dbname).Collection(data_collection_name_prefix)
	_, err := c.InsertOne(context.TODO(), s)
	return err
}

// This function is used in tests only
func (db *Mongodb) replaceRecord(dbname string, id int, s interface{}) error {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}

	c := db.client.Database(dbname).Collection(data_collection_name_prefix)
	_, err := c.ReplaceOne(context.TODO(), bson.M{"message_id": id}, s)
	return err
}

// This function is used in tests only
func (db *Mongodb) insertMeta(dbname string, s interface{}) error {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}

	c := db.client.Database(dbname).Collection(meta_collection_name)
	_, err := c.InsertOne(context.TODO(), s)
	return err
}

func maxIndexQuery(request Request, returnIncompete bool) bson.M {
	var q bson.M
	if request.DatasetOp && !returnIncompete {
		if request.MinDatasetSize > 0 {
			q = bson.M{"$expr": bson.M{"$gte": []interface{}{bson.M{"$size": "$messages"}, request.MinDatasetSize}}, "stream": request.Stream}
		} else {
			q = bson.M{"$expr": bson.M{"$eq": []interface{}{"$size", bson.M{"$size": "$messages"}}}, "stream": request.Stream}
		}
		q = bson.M{"$or": []interface{}{bson.M{"name": finish_stream_keyword, "stream": request.Stream}, q}}
	} else {
		q = bson.M{"stream": request.Stream}
	}
	return q
}

func getInt32Val(request Request, val interface{}) (int32, error) {
	var result int32
	switch v := val.(type) {
	default:
		request.Logger().Debug("unexpected type %T", v)
		return 0, errors.New("cannot convert value to int")
	case int32:
		result, _ = val.(int32)
	case int64:
		val, _ := val.(int64)
		result = int32(val)
	}
	return result, nil
}

func getIntVal(request Request, val interface{}) (int, error) {
	var result int
	switch v := val.(type) {
	default:
		request.Logger().Debug("unexpected type %T", v)
		return 0, errors.New("cannot convert value to int")
	case int32:
		val, _ := val.(int32)
		result = int(val)
	case int64:
		val, _ := val.(int64)
		result = int(val)
	case int:
		val, _ := val.(int)
		result = int(val)
	}
	return result, nil
}

func (db *Mongodb) getMaxIndex(request Request, returnIncompete bool, idKey string) (int, error) {
	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	q := maxIndexQuery(request, returnIncompete)
	opts := options.FindOne().SetSort(bson.M{idKey: -1}).SetProjection(bson.D{{idKey, 1}})
	var result map[string]interface{}
	err := c.FindOne(context.TODO(), q, opts).Decode(&result)
	if err == mongo.ErrNoDocuments {
		return 0, nil
	}
	if err != nil {
		return 0, err
	}
	maxId, ok := getInt32Val(request, result[idKey])
	if ok != nil {
		return 0, errors.New("cannot get max index by " + idKey)
	}
	return int(maxId), nil
}

func duplicateError(err error) bool {
	command_error, ok := err.(mongo.CommandError)
	if !ok {
		write_exception_error, ok1 := err.(mongo.WriteException)
		if !ok1 {
			return false
		}
		return strings.Contains(write_exception_error.Error(), "duplicate key")
	}
	return command_error.Name == "DuplicateKey"
}

func (db *Mongodb) setCounter(request Request, ind int) (err error) {
	update := bson.M{"$set": bson.M{pointer_field_name: ind}}
	opts := options.Update().SetUpsert(true)
	c := db.client.Database(request.DbName()).Collection(pointer_collection_name)
	q := bson.M{"_id": request.GroupId + "_" + request.Stream}
	_, err = c.UpdateOne(context.TODO(), q, update, opts)
	return
}

func (db *Mongodb) errorWhenCannotSetField(request Request, max_ind int) error {
	if res, err := db.getRecordFromDb(request, max_ind, max_ind, "message_id"); err == nil {
		if err2 := checkStreamFinished(request, max_ind, max_ind, res); err2 != nil {
			return err2
		}
	}
	return &DBError{utils.StatusNoData, encodeAnswer(max_ind, max_ind, "")}
}

func message_fallback(res map[string]interface{}) {
	res["_id"] = res["message_id"]
	messages_bson, ok := res["messages"].(primitive.A)
	if !ok {
		return
	}
	var messages_info primitive.A
	for _, msg_bson := range messages_bson {
		msg, _ := msg_bson.(map[string]interface{})
		msg["_id"] = msg["message_id"]
		messages_info = append(messages_info, msg)
	}
	res["messages"] = messages_info
}

func (db *Mongodb) changeField(request Request, change fieldChangeRequest, res interface{}) (err error) {
	var update bson.M
	if change.op == field_op_inc {
		update = bson.M{"$inc": bson.M{change.fieldName: 1}}
	} else if change.op == field_op_set {
		update = bson.M{"$set": bson.M{change.fieldName: change.val}}
	}

	opts := options.FindOneAndUpdate().SetUpsert(true).SetReturnDocument(options.After)
	q := bson.M{"_id": request.GroupId + "_" + request.Stream, change.fieldName: bson.M{"$lt": change.max_ind}}
	c := db.client.Database(request.DbName()).Collection(change.collectionName)

	err = c.FindOneAndUpdate(context.TODO(), q, update, opts).Decode(res)
	if err != nil {
		// duplicateerror can happen because we set Upsert=true to allow insert pointer when it is absent. But then it will try to insert
		// pointer in case query found nothing, also when pointer exists and equal to max_ind. Here we have to return NoData
		if err == mongo.ErrNoDocuments || duplicateError(err) {
			// try again without upsert - if the first error was due to missing pointer
			opts = options.FindOneAndUpdate().SetUpsert(false).SetReturnDocument(options.After)
			if err2 := c.FindOneAndUpdate(context.TODO(), q, update, opts).Decode(res); err2 == nil {
				return nil
			}
			return db.errorWhenCannotSetField(request, change.max_ind)
		}
		return &DBError{utils.StatusTransactionInterrupted, err.Error()}
	}

	return nil
}

func encodeAnswer(id, id_max int, next_stream string) string {
	var r = struct {
		Op          string `json:"op"`
		Id          int    `json:"id"`
		Id_max      int    `json:"id_max"`
		Next_stream string `json:"next_stream"`
	}{"get_record_by_id", id, id_max, next_stream}
	answer, _ := json.Marshal(&r)
	return string(answer)
}

func recordContainsPartialData(request Request, rec map[string]interface{}) bool {
	if !request.DatasetOp {
		return false
	}

	name, ok_name := rec["name"].(string)
	if ok_name && name == finish_stream_keyword {
		return false
	}
	imgs, ok1 := rec["messages"].(primitive.A)
	expectedSize, ok2 := utils.InterfaceToInt64(rec["size"])
	if !ok1 || !ok2 {
		return false
	}
	nMessages := len(imgs)
	if (request.MinDatasetSize == 0 && int64(nMessages) != expectedSize) || (request.MinDatasetSize > 0 && nMessages < request.MinDatasetSize) {
		return true
	}
	return false
}

func addSubstreamSelectionProjection(opts *options.FindOneOptions, substream int) {
	projection := bson.M{
		"messages": bson.M{
			"$elemMatch": bson.M{
				"dataset_substream": substream,
			},
		}, "stream": 1, "message_id": 1, "time_id": 1, "size": 1, "timestamp": 1,
	}
	opts = opts.SetProjection(projection)
}

func (db *Mongodb) getRecordFromDb(request Request, id, id_max int, idKey string) (res map[string]interface{}, err error) {
	q := bson.M{idKey: id, "stream": request.Stream}
	opts := options.FindOne()
	if request.Substream > 0 {
		addSubstreamSelectionProjection(opts, request.Substream)
	}
	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	err = c.FindOne(context.TODO(), q, opts).Decode(&res)
	if err != nil {
		answer := encodeAnswer(id, id_max, "")
		request.Logger().WithFields(map[string]interface{}{"id": id, "cause": err.Error()}).Debug("error getting record")
		return res, &DBError{utils.StatusNoData, answer}
	}
	return res, err
}

func (db *Mongodb) getRecordByIDRaw(request Request, id, id_max int, idKey string) ([]byte, error) {
	res, err := db.getRecordFromDb(request, id, id_max, idKey)
	if err != nil {
		return nil, err
	}

	if err := checkStreamFinished(request, id, id_max, res); err != nil {
		return nil, err
	}

	request.Logger().WithFields(map[string]interface{}{"id": id}).Debug("got record from db by ", idKey)

	message_fallback(res)
	record, err := utils.MapToJson(&res)
	if err != nil {
		return nil, err
	}
	if recordContainsPartialData(request, res) {
		return nil, &DBError{utils.StatusPartialData, string(record)}
	} else {
		return record, nil
	}
}

func (db *Mongodb) getRawRecordWithSort(dbname string, collection_name string, sortField string, sortOrder int) (map[string]interface{}, error) {
	var res map[string]interface{}
	c := db.client.Database(dbname).Collection(data_collection_name_prefix)
	opts := options.FindOne().SetSort(bson.M{sortField: sortOrder})
	var q bson.M = bson.M{"stream": collection_name}
	err := c.FindOne(context.TODO(), q, opts).Decode(&res)

	if err != nil {
		if err == mongo.ErrNoDocuments {
			return map[string]interface{}{}, nil
		}
		return nil, err
	}
	return res, nil
}

func (db *Mongodb) getLastRawRecord(dbname string, collection_name string) (map[string]interface{}, error) {
	return db.getRawRecordWithSort(dbname, collection_name, "message_id", -1)
}

func (db *Mongodb) getEarliestRawRecord(dbname string, collection_name string) (map[string]interface{}, error) {
	return db.getRawRecordWithSort(dbname, collection_name, "timestamp", 1)
}

func (db *Mongodb) getRecordByID(request Request) ([]byte, error, uint64) {
	params, ok := request.ExtraParam.(ExtraParamId)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, failed_extract_extra_params}, 0
	}

	max_ind, err := db.getMaxIndex(request, true, params.IdKey)
	if err != nil {
		return nil, err, 0
	}

	data, err := db.getRecordByIDRaw(request, params.Id, max_ind, params.IdKey)
	return data, err, uint64(params.Id)
}

func (db *Mongodb) negAckRecord(request Request) ([]byte, error) {
	params, ok := request.ExtraParam.(ExtraParamNegAck)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}

	err := db.InsertRecordToInprocess(request, inprocess_collection_name_prefix, params.Id, params.DelayMs, 1, true)
	return []byte(""), err
}

func (db *Mongodb) ackRecord(request Request) ([]byte, error) {
	params, ok := request.ExtraParam.(ExtraParamAck)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}
	askRecord := AckRecord{
		params.Id, request.Stream, request.GroupId,
	}
	c := db.client.Database(request.DbName()).Collection(acks_collection_name_prefix)
	_, err := c.InsertOne(context.TODO(), &askRecord)
	if err != nil {
		if duplicateError(err) {
			return nil, &DBError{utils.StatusWrongInput, "already acknowledged"}
		}
		return nil, err
	}

	c = db.client.Database(request.DbName()).Collection(inprocess_collection_name_prefix)
	filter := bson.M{"message_id": params.Id, "stream": request.Stream, "group_id": request.GroupId}
	_, err_del := c.DeleteOne(context.TODO(), filter)
	if err_del != nil {
		return nil, &DBError{utils.StatusWrongInput, err.Error()}
	}

	return []byte(""), err
}

func (db *Mongodb) checkDatabaseOperationPrerequisites(request Request) error {
	if db.client == nil {
		return &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}

	if len(request.DbName()) <= 1 || len(request.Stream) == 0 {
		return &DBError{utils.StatusWrongInput, "beamtime_id ans stream must be set"}
	}

	return nil
}

func (db *Mongodb) getCurrentPointer(request Request, collectionName string, idKey string) (LocationPointer, int, error) {
	max_ind, err := db.getMaxIndex(request, true, idKey)
	if err != nil {
		return LocationPointer{}, 0, err
	}

	if max_ind == 0 {
		return LocationPointer{}, 0, &DBError{utils.StatusNoData, encodeAnswer(0, 0, "")}
	}

	var curPointer LocationPointer
	err = db.changeField(request, fieldChangeRequest{
		collectionName: collectionName,
		fieldName:      pointer_field_name,
		op:             field_op_inc,
		max_ind:        max_ind}, &curPointer)
	if err != nil {
		return LocationPointer{}, 0, err
	}

	return curPointer, max_ind, nil
}

func (db *Mongodb) getUnProcessedId(dbname string, request Request, collection_name string, delayMs int, nResendAttempts int, rlog log.Logger) (int, error) {
	var res InProcessingRecord
	opts := options.FindOneAndUpdate().SetUpsert(false).SetReturnDocument(options.After)
	tNow := time.Now().UnixNano()
	var update bson.M
	if nResendAttempts == 0 {
		update = bson.M{"$set": bson.M{"delayMs": tNow + int64(delayMs*1e6), "maxResendAttempts": math.MaxInt32}, "$inc": bson.M{"resendAttempts": 1}}
	} else {
		update = bson.M{"$set": bson.M{"delayMs": tNow + int64(delayMs*1e6), "maxResendAttempts": nResendAttempts}, "$inc": bson.M{"resendAttempts": 1}}
	}

	q := bson.M{"stream": request.Stream, "group_id": request.GroupId,
		"delayMs": bson.M{"$lte": tNow},
		"$expr":   bson.M{"$lt": []string{"$resendAttempts", "$maxResendAttempts"}}}
	c := db.client.Database(dbname).Collection(collection_name)
	err := c.FindOneAndUpdate(context.TODO(), q, update, opts).Decode(&res)
	if err != nil {
		if err == mongo.ErrNoDocuments {
			return 0, nil
		}
		return 0, err
	}
	rlog.WithFields(map[string]interface{}{"id": res.Id}).Debug("got unprocessed message")
	return res.Id, nil
}

func (db *Mongodb) InsertRecordToInprocess(request Request, collection_name string, id int, delayMs int, nResendAttempts int, replaceIfExist bool) error {
	record := InProcessingRecord{
		id, request.Stream, request.GroupId,
		nResendAttempts, 0,
		time.Now().UnixNano() + int64(delayMs*1e6),
	}

	c := db.client.Database(request.DbName()).Collection(collection_name)
	_, err := c.InsertOne(context.TODO(), &record)
	if duplicateError(err) {
		if !replaceIfExist {
			return nil
		}
		_, err := c.ReplaceOne(context.TODO(), bson.M{"message_id": id}, &record)
		return err
	}
	return err
}

func (db *Mongodb) InsertToInprocessIfNeeded(request Request, collection_name string, id int, params ExtraParamNext) error {
	if !params.Resend {
		return nil
	}
	return db.InsertRecordToInprocess(request, collection_name, id, params.DelayMs, params.ResendAttempts, false)
}

func (db *Mongodb) getNextAndMaxIndexesFromInprocessed(request Request, params ExtraParamNext, ignoreTimeout bool) (int, int, error) {
	var record_ind, max_ind int
	var err error
	tNow := time.Now().Unix()
	dbSessionLock.Lock()
	t := db.lastReadFromInprocess[request.Stream+"_"+request.GroupId]
	dbSessionLock.Unlock()
	if (t <= tNow-int64(db.settings.ReadFromInprocessPeriod)) || ignoreTimeout {
		record_ind, err = db.getUnProcessedId(request.DbName(), request, inprocess_collection_name_prefix, params.DelayMs, params.ResendAttempts,
			request.Logger())
		if err != nil {
			request.Logger().WithFields(map[string]interface{}{"cause": err.Error()}).Debug("error getting unprocessed message")
			return 0, 0, err
		}
	}
	if record_ind != 0 {
		max_ind, err = db.getMaxIndex(request, true, params.IdKey)
		if err != nil {
			return 0, 0, err
		}
	} else {
		dbSessionLock.Lock()
		db.lastReadFromInprocess[request.Stream+"_"+request.GroupId] = time.Now().Unix()
		dbSessionLock.Unlock()
	}
	return record_ind, max_ind, nil

}

func (db *Mongodb) getNextAndMaxIndexesFromCurPointer(request Request, collectionName, idKey string) (int, int, error) {
	curPointer, max_ind, err := db.getCurrentPointer(request, collectionName, idKey)
	if err != nil {
		request.Logger().Debug("error getting next pointer", err)
		request.Logger().WithFields(map[string]interface{}{"cause": err.Error()}).Debug("error getting next pointer")
		return 0, 0, err
	}
	request.Logger().WithFields(map[string]interface{}{"id": curPointer.Value}).Debug("got next pointer")
	return curPointer.Value, max_ind, nil
}

func (db *Mongodb) getNextAndMaxIndexes(request Request, params ExtraParamNext) (int, int, error) {
	nextInd, maxInd, err := db.getNextAndMaxIndexesFromInprocessed(request, params, false)
	if err != nil {
		return 0, 0, err
	}

	if nextInd == 0 {
		nextInd, maxInd, err = db.getNextAndMaxIndexesFromCurPointer(request, pointer_collection_name, params.IdKey)
		if err_db, ok := err.(*DBError); ok && err_db.Code == utils.StatusNoData {
			var err_inproc error
			nextInd, maxInd, err_inproc = db.getNextAndMaxIndexesFromInprocessed(request, params, true)
			if err_inproc != nil {
				return 0, 0, err_inproc
			}
			if nextInd == 0 {
				return 0, 0, err
			}
		}
	}
	return nextInd, maxInd, nil
}

func ExtractMessageRecord(data map[string]interface{}) (MessageRecord, bool) {
	var r MessageRecord
	err := utils.MapToStruct(data, &r)
	if err != nil {
		return r, false
	}
	r.FinishedStream = (r.Name == finish_stream_keyword)
	if r.FinishedStream {
		var next_stream string
		next_stream, ok := r.Meta["next_stream"].(string)
		if !ok {
			next_stream = no_next_stream_keyword
		}
		r.NextStream = next_stream
	}
	return r, true
}

//Return next in time (based on current pointer) available record from DB
func (db *Mongodb) getNextRecordDB(request Request, currentPointer int) (map[string]interface{}, error) {
	var res map[string]interface{}
	filter := bson.M{"time_id": bson.M{"$gt": currentPointer}, "stream": request.Stream}
	opts := options.FindOne().SetSort(bson.M{"time_id": 1})
	if request.Substream > 0 {
		addSubstreamSelectionProjection(opts, request.Substream)
	}
	coll := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	err := coll.FindOne(context.TODO(), filter, opts).Decode(&res)
	if err != nil {
		return res, &DBError{utils.StatusNoData, err.Error()}
	}
	return res, nil
}

func (db *Mongodb) getCurrentPointerDB(request Request) (LocationPointer, error) {
	var curPointer LocationPointer
	filter := bson.M{"_id": request.GroupId + "_" + request.Stream}
	opts := options.FindOne()
	coll := db.client.Database(request.DbName()).Collection(pointer_collection_name)
	err := coll.FindOne(context.TODO(), filter, opts).Decode(&curPointer)
	if err != nil {
		return curPointer, &DBError{utils.StatusTransactionInterrupted, err.Error()}
	}
	return curPointer, nil
}

// If value of the current pointer didn't change during processing of getNext request
// update it to next existing record.
func (db *Mongodb) updateCurrentPointerIfEqualDB(request Request, currentValue, nextValue int) error {
	update := bson.M{"$set": bson.M{pointer_field_name: nextValue}}
	opts := options.Update().SetUpsert(false)
	coll := db.client.Database(request.DbName()).Collection(pointer_collection_name)
	filter := bson.M{"_id": request.GroupId + "_" + request.Stream, pointer_field_name: currentValue}
	_, err := coll.UpdateOne(context.TODO(), filter, update, opts)
	return err
}

// Return record based on current pointer, index of the record and error
// Return actual value of _id, even if request is done for time_id
// If obtaining current pointer fails, record is corrupted or thisis end of stream: return 0 instead of record id
// If record is missing, or incomplete return valid record id
func (db *Mongodb) getRecordFromCurPointer(request Request, params ExtraParamNext) ([]byte, int, error) {

	nextInd, maxInd, err := db.getNextAndMaxIndexesFromCurPointer(request, pointer_collection_name, params.IdKey)
	if err != nil {
		return nil, 0, err
	}
	var res map[string]interface{}
	res, err = db.getRecordFromDb(request, nextInd, maxInd, params.IdKey)
	var idx int
	if params.IdKey == "message_id" {
		idx = nextInd
	}
	if err == nil && params.IdKey == "time_id" {
		idx, _ = getIntVal(request, res["message_id"])
	}
	// Missing Id! Since time_Ids are sequential, jump to the next available id
	// Update current pointer to next available id
	if err != nil && params.IdKey == "time_id" {
		for {
			curPointer, err := db.getCurrentPointerDB(request)
			if err != nil {
				return nil, 0, err
			}
			res, err = db.getNextRecordDB(request, curPointer.Value)
			if err != nil {
				return nil, 0, err
			}
			nextValue, ok := getInt32Val(request, res["time_id"])
			if ok != nil {
				return nil, 0, errors.New("failed to get next available id. Extraction of id fails")
			}
			idx, ok = getIntVal(request, res["message_id"])
			if ok != nil {
				return nil, 0, errors.New("failed to get _id. Extraction of _id fails")
			}
			updateErr := db.updateCurrentPointerIfEqualDB(request, curPointer.Value, int(nextValue))
			// If update fails, another broker already change the value. Go to the next one.
			if updateErr == nil {
				nextInd = int(nextValue)
				break
			}
		}
	} else {
		// In this case we can not guess, what is the next
		if err != nil {
			return nil, idx, err
		}
	}

	if err := checkStreamFinished(request, nextInd, maxInd, res); err != nil {
		return nil, 0, err
	}

	request.Logger().WithFields(map[string]interface{}{"id": idx}).Debug("got record from db by id key: ",
		params.IdKey)

	message_fallback(res)
	record, err := utils.MapToJson(&res)
	if err != nil {
		return nil, 0, err
	}
	if recordContainsPartialData(request, res) {
		return nil, idx, &DBError{utils.StatusPartialData, string(record)}
	} else {
		return record, idx, nil
	}

}

// Get index of the record from in-process collection and
// return corresponding record
func (db *Mongodb) getRecordFromInprocessed(request Request, params ExtraParamNext,
	originalerror error, ignoreTimeout bool) ([]byte, int, error) {
	var err_inproc error
	nextInd, maxInd, err_inproc := db.getNextAndMaxIndexesFromInprocessed(request, params, ignoreTimeout)
	if err_inproc != nil {
		return nil, 0, err_inproc
	}
	if nextInd != 0 {
		data, err := db.getRecordByIDRaw(request, nextInd, maxInd, params.IdKey)
		return data, nextInd, err
	} else {
		return nil, 0, originalerror
	}
}

func checkStreamFinished(request Request, id, id_max int, data map[string]interface{}) error {
	if id != id_max {
		return nil
	}
	r, ok := ExtractMessageRecord(data)
	if !ok || !r.FinishedStream {
		return nil
	}
	request.Logger().WithFields(map[string]interface{}{"nextStream": r.NextStream}).Debug("reached end of stream")

	answer := encodeAnswer(r.ID-1, r.ID-1, r.NextStream)
	return &DBError{utils.StatusNoData, answer}
}

func (db *Mongodb) getNextRecord(request Request) ([]byte, error, uint64) {
	params, ok := request.ExtraParam.(ExtraParamNext)
	if !ok {
		return nil, errors.New(failed_extract_extra_params), 0
	}
	request.Logger().Debug("get next by : ", params.IdKey)

	var data []byte
	var nextInd int
	data, nextInd, err := db.getRecordFromInprocessed(request, params, nil, false)
	if err != nil {
		return nil, err, 0
	}

	missingInd := 0
	if data == nil {
		data, nextInd, err = db.getRecordFromCurPointer(request, params)
		if err != nil {
			request.Logger().Debug("error getting record from current pointer: ", err)
			missingInd = nextInd
		}
	}

	if data == nil {
		data, nextInd, err = db.getRecordFromInprocessed(request, params, err, true)
	}

	// If index of the record is valid, store it in in-process table to try again next time
	if missingInd > 0 {
		errUpdate := db.InsertToInprocessIfNeeded(request,
			inprocess_collection_name_prefix, missingInd, params)
		if errUpdate != nil {
			return nil, errUpdate, 0
		}
	}

	if err == nil {
		err_update := db.InsertToInprocessIfNeeded(request,
			inprocess_collection_name_prefix, nextInd, params)
		if err_update != nil {
			return nil, err_update, 0
		}
	}
	return data, err, uint64(nextInd)
}

func (db *Mongodb) getLastRecord(request Request, idKey string) ([]byte, error, uint64) {
	max_ind, err := db.getMaxIndex(request, false, idKey)
	if err != nil {
		return nil, err, 0
	}
	request.Logger().WithFields(map[string]interface{}{"id": max_ind}).Debug("Get max index by " + idKey)
	data, err := db.getRecordByIDRaw(request, max_ind, max_ind, idKey)
	return data, err, uint64(max_ind)
}

func (db *Mongodb) getLastRecordInGroup(request Request, idKey string) ([]byte, error) {
	max_ind, err := db.getMaxIndex(request, false, idKey)
	if err != nil {
		return nil, err
	}

	var res map[string]interface{}
	err = db.changeField(request, fieldChangeRequest{
		collectionName: last_message_collection_name,
		fieldName:      last_message_field_name,
		op:             field_op_set,
		max_ind:        max_ind,
		val:            max_ind,
	}, &res)
	if err != nil {
		return nil, err
	}
	return db.getRecordByIDRaw(request, max_ind, max_ind, idKey)
}

func getSizeFilter(request Request) bson.M {
	filter := bson.M{"stream": request.Stream}
	params := request.ExtraParam.(ExtraParamSize)
	if params.IsDataset {
		if params.IncompleteDataset {
			filter["$expr"] = bson.M{"$gt": []interface{}{bson.M{"$size": "$messages"}, 0}}
		} else {
			filter["$expr"] = bson.M{"$eq": []interface{}{"$size", bson.M{"$size": "$messages"}}}
		}
	}
	filter["name"] = bson.M{"$ne": finish_stream_keyword}
	return filter
}

// Calculate number of messages for given stream
func (db *Mongodb) getSize(request Request) ([]byte, error) {
	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)

	filter := getSizeFilter(request)
	size, err := c.CountDocuments(context.TODO(), filter, options.Count())
	if err != nil {
		if ce, ok := err.(mongo.CommandError); ok && ce.Code == 17124 {
			return nil, &DBError{utils.StatusWrongInput, "no datasets found"}
		}
		return nil, err
	}

	var rec SizeRecord
	rec.Size = int(size)
	return json.Marshal(&rec)
}

func (db *Mongodb) resetCounter(request Request) ([]byte, error) {
	id, ok := request.ExtraParam.(int)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}

	err := db.setCounter(request, id)
	if err != nil {
		return []byte(""), err
	}

	c := db.client.Database(request.DbName()).Collection(inprocess_collection_name_prefix)
	filter := bson.M{"message_id": bson.M{"$gte": id}, "stream": request.Stream,
		"group_id": request.GroupId}
	_, err_del := c.DeleteMany(context.TODO(), filter)
	if err_del != nil {
		return nil, &DBError{utils.StatusWrongInput, err_del.Error()}
	}

	return []byte(""), nil
}

func getMetaId(request Request) (string, error) {
	meta_id, ok := request.ExtraParam.(string)
	if !ok {
		return "", &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}
	switch meta_id {
	case "0":
		return "bt", nil
	case "1":
		return "st_" + request.Stream, nil
	default:
		return "", &DBError{utils.StatusWrongInput, "wrong meta type"}
	}
}

func (db *Mongodb) getMeta(request Request) ([]byte, error) {
	id, err := getMetaId(request)
	if err != nil {
		return nil, err
	}
	q := bson.M{"_id": id}
	var res map[string]interface{}
	c := db.client.Database(request.DbName()).Collection(meta_collection_name)
	err = c.FindOne(context.TODO(), q, options.FindOne()).Decode(&res)
	if err != nil {
		request.Logger().WithFields(map[string]interface{}{"meta_id": id, "cause": err.Error()}).Debug("error getting meta")
		return nil, &DBError{utils.StatusNoData, err.Error()}
	}
	userMeta, ok := res["meta"]
	if !ok {
		request.Logger().WithFields(map[string]interface{}{"id": id, "cause": "cannot parse database response"}).Debug("error getting meta")
		return nil, errors.New("cannot get metadata")
	}
	request.Logger().WithFields(map[string]interface{}{"meta_id": id}).Error("got metadata")
	return utils.MapToJson(&userMeta)
}

func (db *Mongodb) processQueryError(query, dbname string, err error, rlog log.Logger) ([]byte, error) {
	rlog.WithFields(map[string]interface{}{"query": query, "cause": err.Error()}).Debug("error processing query")
	return nil, &DBError{utils.StatusNoData, err.Error()}
}

func (db *Mongodb) queryMessages(request Request) ([]byte, error) {
	var res []map[string]interface{}
	query := "stream = '" + request.Stream + "' AND " + request.ExtraParam.(string)
	q, sort, err := db.BSONFromSQL(request.DbName(), query)
	if err != nil {
		request.Logger().WithFields(map[string]interface{}{"query": request.ExtraParam.(string), "cause": err.Error()}).Debug("error parsing query")
		return nil, &DBError{utils.StatusWrongInput, err.Error()}
	}

	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	opts := options.Find()

	if len(sort) > 0 {
		opts = opts.SetSort(sort)
	} else {
	}

	cursor, err := c.Find(context.TODO(), q, opts)
	if err != nil {
		return db.processQueryError(query, request.DbName(), err, request.Logger())
	}
	err = cursor.All(context.TODO(), &res)
	if err != nil {
		return db.processQueryError(query, request.DbName(), err, request.Logger())
	}

	request.Logger().WithFields(map[string]interface{}{"query": request.ExtraParam.(string), "recordsFound": len(res)}).Debug("processed query")

	for _, msg := range res {
		message_fallback(msg)
	}
	if res != nil {
		return utils.MapToJson(&res)
	} else {
		return []byte("[]"), nil
	}
}

func makeRange(min, max int) []int {
	a := make([]int, max-min+1)
	for i := range a {
		a[i] = min + i
	}
	return a
}

func (db *Mongodb) getNacksLimits(request Request) (int, int, error) {
	pars, ok := request.ExtraParam.(ExtraParamNacks)
	if !ok {
		return 0, 0, &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}

	var err error
	if pars.To == 0 {
		pars.To, err = db.getMaxLimitWithoutEndOfStream(request, "message_id")
		if err != nil {
			return 0, 0, err
		}
	}
	return pars.From, pars.To, nil
}

func (db *Mongodb) getMaxLimitWithoutEndOfStream(request Request, idKey string) (int, error) {
	maxInd, err := db.getMaxIndex(request, true, idKey)
	if err != nil {
		return 0, err
	}
	_, last_err := db.getRecordByIDRaw(request, maxInd, maxInd, idKey)
	if last_err != nil && maxInd > 0 {
		maxInd = maxInd - 1
	}
	return maxInd, nil
}

func (db *Mongodb) nacks(request Request) ([]byte, error) {
	from, to, err := db.getNacksLimits(request)
	if err != nil {
		return nil, err
	}

	res := Nacks{[]int{}}
	if to == 0 {
		return utils.MapToJson(&res)
	}

	res.Unacknowledged, err = db.getNacks(request, from, to)
	if err != nil {
		return nil, err
	}

	return utils.MapToJson(&res)
}

func (db *Mongodb) deleteCollection(request Request, name string) error {
	return db.client.Database(request.DbName()).Collection(name).Drop(context.TODO())
}

func (db *Mongodb) collectionExist(request Request, name string) (bool, error) {
	result, err := db.client.Database(request.DbName()).ListCollectionNames(context.TODO(), bson.M{"name": name})
	if err != nil {
		return false, err
	}
	if len(result) == 1 {
		return true, nil
	}
	return false, nil
}

func (db *Mongodb) deleteDataCollection(errorOnNotexist bool, request Request) error {
	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	filter := bson.M{"stream": request.Stream}
	count, err := c.DeleteMany(context.TODO(), filter)
	if err != nil {
		return &DBError{utils.StatusWrongInput, err.Error()}
	}
	if count.DeletedCount == 0 && errorOnNotexist {
		return &DBError{utils.StatusWrongInput, "No messages in given stream to be deleted"}
	}
	return nil
}

func (db *Mongodb) deleteDocumentsInCollection(request Request, collection string, field string, pattern string) error {
	filter := bson.M{field: bson.D{{"$regex", primitive.Regex{Pattern: pattern, Options: "i"}}}}
	_, err := db.client.Database(request.DbName()).Collection(collection).DeleteMany(context.TODO(), filter)
	return err
}

func escapeQuery(query string) (res string) {
	chars := `\-[]{}()*+?.,^$|#`
	for _, char := range chars {
		query = strings.ReplaceAll(query, string(char), `\`+string(char))
	}
	return query
}

func (db *Mongodb) deleteCollectionsWithPrefix(request Request, prefix string) error {
	cols, err := db.client.Database(request.DbName()).ListCollectionNames(context.TODO(), bson.M{"name": bson.D{
		{"$regex", primitive.Regex{Pattern: "^" + escapeQuery(prefix), Options: "i"}}}})
	if err != nil {
		return err
	}

	for _, col := range cols {
		err := db.deleteCollection(request, col)
		if err != nil {
			return err
		}
	}

	return nil
}

func (db *Mongodb) deleteServiceMeta(request Request) error {
	// Delete InProcess
	c := db.client.Database(request.DbName()).Collection(inprocess_collection_name_prefix)
	filter := bson.M{"stream": request.Stream}
	_, err := c.DeleteMany(context.TODO(), filter)
	if err != nil {
		return err
	}

	// Delete acks
	c = db.client.Database(request.DbName()).Collection(acks_collection_name_prefix)
	filter = bson.M{"stream": request.Stream}
	_, err = c.DeleteMany(context.TODO(), filter)
	if err != nil {
		return err
	}

	// Delete current location
	err = db.deleteDocumentsInCollection(request, pointer_collection_name, "_id", ".*_"+escapeQuery(request.Stream))
	if err != nil {
		return err
	}

	// Delete meta
	err = db.deleteDocumentsInCollection(request, meta_collection_name, "_id", ".*_"+escapeQuery(request.Stream)+"_.*")
	if err != nil {
		return err
	}

	// Delete from stream list
	err = db.deleteDocumentsInCollection(request, streams_info, "name", escapeQuery(request.Stream))
	if err != nil {
		return err
	}
	return nil
}

func (db *Mongodb) deleteStream(request Request) ([]byte, error) {
	params, ok := request.ExtraParam.(ExtraParamDelete)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, failed_extract_extra_params}
	}

	if !params.DeleteMeta {
		request.Logger().Debug("skipping delete stream meta")
		return nil, nil
	}

	err := db.deleteDataCollection(params.ErrorOnNotExist, request)
	if err != nil {
		return nil, err
	}

	delete(streams.records, request.DbName())

	err = db.deleteServiceMeta(request)
	return nil, err
}

func (db *Mongodb) persistStream(request Request) ([]byte, error) {
	if db.client == nil {
		return nil, &DBError{utils.StatusServiceUnavailable, no_session_msg}
	}

	maxNum, _ := request.ExtraParam.(int)

	c := db.client.Database(request.DbName()).Collection(meta_collection_name)
	q := bson.M{"_id": "persist"}
	l := PersistedStreamsList{}
	err := c.FindOne(context.TODO(), q, options.FindOne()).Decode(&l)
	if err != nil && err != mongo.ErrNoDocuments {
		return nil, err
	}
	if len(l.Streams) >= maxNum {
		return nil, &DBError{utils.StatusWrongInput, too_many_persisted_streams}
	}
	l.Streams = append(l.Streams, request.Stream)

	_, err = c.InsertOne(context.TODO(), l)

	return nil, err
}

func (db *Mongodb) lastAck(request Request) ([]byte, error) {
	c := db.client.Database(request.DbName()).Collection(acks_collection_name_prefix)
	opts := options.FindOne().SetSort(bson.M{"message_id": -1}).SetReturnKey(true)
	result := LastAck{0}
	q := bson.M{"stream": request.Stream, "group_id": request.GroupId}
	err := c.FindOne(context.TODO(), q, opts).Decode(&result)
	if err == mongo.ErrNoDocuments {
		return utils.MapToJson(&result)
	}
	if err != nil {
		return nil, err
	}

	return utils.MapToJson(&result)
}

func (db *Mongodb) canAvoidDbRequest(min_index int, max_index int, c *mongo.Collection) ([]int, error, bool) {
	if min_index > max_index {
		return []int{}, errors.New("from index is greater than to index"), true
	}

	size, err := c.CountDocuments(context.TODO(), bson.M{}, options.Count())
	if err != nil {
		return []int{}, err, true
	}

	if size == 0 {
		return makeRange(min_index, max_index), nil, true
	}

	if min_index == 1 && int(size) == max_index {
		return []int{}, nil, true
	}

	return nil, nil, false
}

func getNacksQuery(request Request, max_index int, min_index int) []bson.D {
	matchStage := bson.D{
		{"$match", bson.D{
			{"message_id", bson.D{
				{"$lt", max_index + 1},
				{"$gt", min_index - 1},
			}},
			{"stream", request.Stream},
			{"group_id", request.GroupId},
		}},
	}

	groupStage := bson.D{
		{"$group", bson.D{
			{"_id", 0},
			{"numbers", bson.D{
				{"$push", "$message_id"},
			}}},
		}}
	projectStage := bson.D{
		{"$project", bson.D{
			{"_id", 0},
			{"numbers", bson.D{
				{"$setDifference", bson.A{bson.D{{"$range", bson.A{min_index, max_index + 1}}}, "$numbers"}},
			}}},
		}}
	return mongo.Pipeline{matchStage, groupStage, projectStage}
}

func extractNacsFromCursor(err error, cursor *mongo.Cursor) ([]int, error) {
	resp := []struct {
		Numbers []int
	}{}
	err = cursor.All(context.TODO(), &resp)
	if err != nil || len(resp) != 1 {
		return []int{}, err
	}
	return resp[0].Numbers, nil
}

func (db *Mongodb) getNacks(request Request, min_index, max_index int) ([]int, error) {
	c := db.client.Database(request.DbName()).Collection(acks_collection_name_prefix)

	if res, err, ok := db.canAvoidDbRequest(min_index, max_index, c); ok {
		return res, err
	}

	query := getNacksQuery(request, max_index, min_index)
	cursor, err := c.Aggregate(context.TODO(), query)
	if err != nil {
		return []int{}, err
	}
	return extractNacsFromCursor(err, cursor)
}

func (db *Mongodb) getStreams(request Request) ([]byte, error) {
	rec, err := streams.getStreams(db, request)
	if err != nil {
		return db.processQueryError("get streams", request.DbName(), err, request.Logger())
	}
	return json.Marshal(&rec)
}

func (db *Mongodb) getDatasources(request Request) ([]byte, error) {

	filter := bson.M{"name": primitive.Regex{Pattern: "^" + request.Beamtime + "_"}}
	rec, err := db.client.ListDatabaseNames(context.TODO(), filter)
	if err != nil {
		return db.processQueryError("get streams", request.DbName(), err, request.Logger())
	}
	for i, source := range rec {
		rec[i] = source[len(request.Beamtime)+1:]
	}
	res := Sources{rec}
	return json.Marshal(&res)
}

func (db *Mongodb) ProcessRequest(request Request) (answer []byte, err error, idx uint64) {
	dbClientLock.RLock()
	defer dbClientLock.RUnlock()

	if err := db.checkDatabaseOperationPrerequisites(request); err != nil {
		return nil, err, 0
	}

	if err := encodeRequest(&request); err != nil {
		return nil, err, 0
	}

	idx = 0
	switch request.Op {
	case "next":
		answer, err, idx = db.getNextRecord(request)
	case "id":
		answer, err, idx = db.getRecordByID(request)
	case "last":
		answer, err, idx = db.getLastRecord(request, "message_id")
	case "groupedlast":
		answer, err = db.getLastRecordInGroup(request, "message_id")
	case "resetcounter":
		answer, err = db.resetCounter(request)
	case "size":
		answer, err = db.getSize(request)
	case "meta":
		answer, err = db.getMeta(request)
	case "querymessages":
		answer, err = db.queryMessages(request)
	case "streams":
		answer, err = db.getStreams(request)
	case "datasources":
		answer, err = db.getDatasources(request)
	case "ackmessage":
		answer, err = db.ackRecord(request)
	case "negackmessage":
		answer, err = db.negAckRecord(request)
	case "nacks":
		answer, err = db.nacks(request)
	case "lastack":
		answer, err = db.lastAck(request)
	case "delete_stream":
		answer, err = db.deleteStream(request)
	case "persist_stream":
		answer, err = db.persistStream(request)
	case "write_meta":
		answer, err = db.writeMeta(request)
	case "write_message":
		answer, err = db.writeMessage(request)
	default:
		answer, err = nil, errors.New("Wrong db operation: "+request.Op)
	}

	return answer, err, idx
}
