package database

import (
	"asapo_common/utils"
	"context"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
)

type ExtraParamMessage struct {
	Id          int64      `json:"id"`
	DatasetSize int        `json:"dataset_size"`
	Message     MessageRaw `json:"message"`
}

type ExtraParamMeta struct {
	Id      string                 `json:"id"`
	Mode    string                 `json:"mode"`
	Upsert  bool                   `json:"upsert"`
	Message map[string]interface{} `json:"message"`
}

type MessageRaw struct {
	Id               int64                  `bson:"message_id" json:"_id"`
	TimeId           int64                  `bson:"time_id" json:"time_id"`
	Timestamp        int64                  `bson:"timestamp" json:"timestamp"`
	Name             string                 `bson:"name" json:"name"`
	Meta             map[string]interface{} `bson:"meta" json:"meta"`
	Source           string                 `bson:"source" json:"source"`
	IbSource         string                 `bson:"ib_source" json:"ib_source"`
	BufId            int64                  `bson:"buf_id" json:"buf_id"`
	Stream           string                 `bson:"stream" json:"stream"`
	DatasetSubstream int                    `bson:"dataset_substream" json:"dataset_substream"`
	IngestMode       int32                  `bson:"ingest_mode" json:"ingest_mode"`
	Size             int32                  `bson:"size" json:"size"`
}

func (db *Mongodb) getNextId(request Request) (int64, error) {

	filter := bson.M{"_id": request.DataSource}
	update := bson.M{"$inc": bson.M{"value": int64(1)}}
	options := options.FindOneAndUpdate().SetUpsert(true).SetReturnDocument(options.After)

	type NextId struct {
		DataSource string `bson:"_id" `
		Id         int64  `bson:"value"`
	}
	var res NextId
	c := db.client.Database(request.DbName()).Collection(auto_id_counter_name)
	err := c.FindOneAndUpdate(context.TODO(), filter, update, options).Decode(&res)
	if err != nil {
		return 0, err
	}

	return res.Id, nil
}

func (db *Mongodb) createIndices(request Request) error {

	indexModel := mongo.IndexModel{
		Keys:    bson.D{{Key: "message_id", Value: 1}, {Key: "stream", Value: 1}}, // Compound index on message_id and stream
		Options: options.Index().SetUnique(true),
	}
	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	indexName, err := c.Indexes().CreateOne(context.TODO(), indexModel)
	if err != nil {
		return &DBError{utils.StatusServiceUnavailable, "failed to create index: " + err.Error()}
	}
	request.Logger().Debug("Create index ", indexName)

	indexModel = mongo.IndexModel{
		Keys:    bson.D{{Key: "message_id", Value: 1}, {Key: "stream", Value: 1}, {Key: "group_id", Value: 1}},
		Options: options.Index().SetUnique(true),
	}
	c = db.client.Database(request.DbName()).Collection(inprocess_collection_name_prefix)
	indexName, err = c.Indexes().CreateOne(context.TODO(), indexModel)
	if err != nil {
		return &DBError{utils.StatusServiceUnavailable, "failed to create index: " + err.Error()}
	}
	request.Logger().Debug("Create index ", indexName)

	c = db.client.Database(request.DbName()).Collection(acks_collection_name_prefix)
	indexName, err = c.Indexes().CreateOne(context.TODO(), indexModel)
	if err != nil {
		return &DBError{utils.StatusServiceUnavailable, "failed to create index: " + err.Error()}
	}
	request.Logger().Debug("Create index ", indexName)

	return nil
}

func (db *Mongodb) writeMessage(request Request) ([]byte, error) {

	params, ok := request.ExtraParam.(ExtraParamMessage)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, "failed to parse extra parameters"}
	}

	// ToDo cache information if index is created
	err := db.createIndices(request)
	if err != nil {
		return nil, err
	}

	timeId, err := db.getNextId(request)
	if err != nil {
		return nil, &DBError{utils.StatusWrongInput, "failed to get next id: " + err.Error()}
	}
	params.Message.TimeId = timeId

	if params.DatasetSize > 0 {
		return db.writeAsDataset(request, params.Message, params.DatasetSize)
	} else {
		return db.writeAsMessage(request, params.Message)
	}
}

func (db *Mongodb) writeAsMessage(request Request, message MessageRaw) ([]byte, error) {

	// AutoId case
	if message.Id == 0 {
		message.Id = message.TimeId
	}

	if message.DatasetSubstream > 0 {
		return nil, &DBError{utils.StatusWrongInput, "DatasetSubstream > 0"}
	}

	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	_, err := c.InsertOne(context.TODO(), message)
	if err != nil {
		request.Logger().WithFields(map[string]interface{}{"id": message.Id, "cause": err.Error()}).Error("failed to insert record")
		return nil, &DBError{utils.StatusWrongInput, "failed to insert record " + err.Error()}
	}
	request.Logger().WithFields(map[string]interface{}{"id": message.Id}).Debug("insert record into database")
	return nil, nil

}

func (db *Mongodb) writeAsDataset(request Request, message MessageRaw, size int) ([]byte, error) {

	filter := bson.M{"message_id": message.Id, "stream": message.Stream, "size": int64(size),
		"messages.dataset_substream": bson.M{"$ne": message.DatasetSubstream}}
	update := bson.M{
		"$set":         bson.M{"timestamp": message.Timestamp},
		"$setOnInsert": bson.M{"time_id": message.TimeId},
		"$push":        bson.M{"messages": message},
	}
	options := options.Update().SetUpsert(true)

	c := db.client.Database(request.DbName()).Collection(data_collection_name_prefix)
	_, err := c.UpdateOne(context.TODO(), filter, update, options)
	// ToDo Commented code was used in receiver to ingest dataset. Check, that it works without it.
	// first update may fail due to multiple threads try to create document at once, the second one should succeed
	// https://jira.mongodb.org/browse/SERVER-14322
	//if duplicateError(err) {
	//	_, err = c.UpdateOne(context.TODO(), filter, update, options)
	//}
	if err != nil {
		request.Logger().WithFields(map[string]interface{}{"id": message.Id,
			"substream": message.DatasetSubstream, "cause": err.Error()}).Error("failed to insert dataset")
		return nil, &DBError{utils.StatusWrongInput, "failed to insert dataset " + err.Error()}
	}
	request.Logger().WithFields(map[string]interface{}{"id": message.Id, "substream": message.DatasetSubstream}).Debug("insert substream record into database")
	return nil, nil
}

func (db *Mongodb) writeMeta(request Request) ([]byte, error) {

	var params, ok = request.ExtraParam.(ExtraParamMeta)
	if !ok {
		return nil, &DBError{utils.StatusWrongInput, "failed to parse extra parameters "}
	}

	c := db.client.Database(request.DbName()).Collection(meta_collection_name)
	// Ingest modes are defined in: common/cpp/include/asapo/common/data_structs.h
	var err error
	switch params.Mode {
	default:
		return nil, &DBError{utils.StatusWrongInput, "Unexpected metadata injest mode " + params.Mode}
	case "1": // Insert
		insert_info := bson.M{"_id": params.Id, "meta": params.Message}
		_, err = c.InsertOne(context.TODO(), insert_info)
	case "2": // Replace
		filter := bson.M{"_id": params.Id}
		update := bson.M{"$set": bson.M{"meta": params.Message}}
		options := options.Update().SetUpsert(params.Upsert)
		_, err = c.UpdateOne(context.TODO(), filter, update, options)
	case "3": // Update
		filter := bson.M{"_id": params.Id}
		update_info := bson.M{}
		for k, v := range params.Message {
			update_info["meta."+k] = v
		}
		update := bson.M{"$set": update_info}
		options := options.Update().SetUpsert(params.Upsert)
		_, err = c.UpdateOne(context.TODO(), filter, update, options)
	}
	if err != nil {
		request.Logger().WithFields(map[string]interface{}{"meta_id": params.Id, "cause": err.Error()}).Error("failed to insert meta")
		return nil, &DBError{utils.StatusWrongInput, "failed to insert meta: " + err.Error()}
	}
	request.Logger().WithFields(map[string]interface{}{"meta_id": params.Id}).Debug("Insert meta into database")
	return nil, err
}
