//go:build !test
// +build !test

package database

import (
	log "asapo_common/logger"
	"asapo_common/utils"
	"context"
	"errors"
	"fmt"
	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"
	"sort"
	"sync"
	"time"
)

type GetStreamsParams struct {
	From     string `json:"from"`     // Filter streams with given name and all later streams. Compatible with Single.
	Filter   string `json:"filter"`   // Filter by status: finished/unfinished/all
	Detailed bool   `json:"detailed"` // Update stream information with the last message in the streams(TimestampLast, LastID, StreamFinishedFlag)
	Last     bool   `json:"last"`     // Filter only last stream from the list of streams. Single flag is ignored.
	Single   bool   `json:"single"`   // Filter only one First stream from the list
}

type StreamInfo struct {
	LastId        int64  `json:"lastId"`
	Name          string `json:"name"`
	Timestamp     int64  `json:"timestampCreated"`
	TimestampLast int64  `json:"timestampLast"`
	Finished      bool   `json:"finished"`
	NextStream    string `json:"nextStream"`
}

type StreamsRecord struct {
	Streams []StreamInfo `json:"streams"`
}

type Streams struct {
	records     map[string]StreamsRecord
	lastUpdated map[string]time.Time
}

var streams = Streams{lastUpdated: make(map[string]time.Time, 0), records: make(map[string]StreamsRecord, 0)}
var streamsLock sync.Mutex

func (ss *Streams) getStreamsInfoFromCache(db_name string, updatePeriodMs int) (StreamsRecord, error) {
	if time.Now().Sub(ss.lastUpdated[db_name]).Milliseconds() >= int64(updatePeriodMs) {
		return StreamsRecord{}, errors.New("cache expired")
	}
	rec, ok := ss.records[db_name]
	if !ok {
		return StreamsRecord{}, errors.New("no records for " + db_name)
	}
	res := StreamsRecord{}
	utils.DeepCopy(rec, &res)
	return res, nil
}

// Return StreamsRecord with names of streams filled from DB
// Function query list of collections for given db_name and them filter it by name
func readAllStreamNames(db *Mongodb, db_name string) (StreamsRecord, error) {
	collection := db.client.Database(db_name).Collection(data_collection_name_prefix)
	result, err := collection.Distinct(context.TODO(), "stream", bson.M{})
	if err != nil {
		return StreamsRecord{}, err
	}
	var rec = StreamsRecord{[]StreamInfo{}}
	for _, coll := range result {
		si := StreamInfo{Name: fmt.Sprint(coll)}
		rec.Streams = append(rec.Streams, si)
	}
	return rec, nil
}

// Read stream info for all streams from streams_info collection
// If correction is empty or does not exist return empty array
// and create collecton with unique index `name`
func readStreamsInfo(db *Mongodb, db_name string) ([]StreamInfo, error) {
	coll := db.client.Database(db_name).Collection(streams_info)
	opts := options.Find().SetSort(bson.D{{"name", 1}}) // Order by name to simplify search
	cursor, err := coll.Find(context.TODO(), bson.D{}, opts)
	if err != nil {
		log.Error("Get streams fails", err)
		return []StreamInfo{}, err
	}
	var rec []StreamInfo
	if err = cursor.All(context.TODO(), &rec); err != nil {
		log.Error("Decoding streams fails", err)
		return []StreamInfo{}, err
	}
	if len(rec) == 0 {
		createIndexStreamInfoDB(db, db_name)
	}
	return rec, nil
}

func findStreamAmongCurrent(currentStreams []StreamInfo, record StreamInfo) (int, bool) {
	// currentStreams is already ordered by Name
	ind := sort.Search(len(currentStreams), func(i int) bool {
		return currentStreams[i].Name >= record.Name
	})
	if ind < len(currentStreams) && currentStreams[ind].Name == record.Name {
		return ind, true
	}
	return -1, false
}

func fillInfoFromEarliestRecord(db *Mongodb, db_name string, rec *StreamsRecord, record StreamInfo, i int) error {
	res, err := db.getEarliestRawRecord(db_name, record.Name)
	if err != nil {
		return err
	}
	ts, ok := utils.GetInt64FromMap(res, "timestamp")
	if ok {
		rec.Streams[i].Timestamp = ts
	} else {
		return errors.New("fillInfoFromEarliestRecord: cannot extact timestamp")
	}
	return nil
}

func fillInfoFromLastRecord(db *Mongodb, db_name string, rec *StreamsRecord, record StreamInfo, i int) error {
	res, err := db.getLastRawRecord(db_name, record.Name)
	if err != nil {
		return err
	}
	mrec, ok := ExtractMessageRecord(res)
	if !ok {
		return errors.New("fillInfoFromLastRecord: cannot extract record")
	}

	rec.Streams[i].LastId = int64(mrec.ID)
	rec.Streams[i].TimestampLast = int64(mrec.Timestamp)
	rec.Streams[i].Finished = mrec.FinishedStream
	if mrec.FinishedStream {
		rec.Streams[i].LastId = rec.Streams[i].LastId - 1
		if mrec.NextStream != no_next_stream_keyword {
			rec.Streams[i].NextStream = mrec.NextStream
		}
	}
	return nil
}

// If stream name from record is found in currentStreams corresponding
// record get copied to rec.
func updateStreamInfoFromCurrent(currentStreams []StreamInfo, record StreamInfo, rec *StreamInfo) (found bool, updateFinished bool) {
	ind, found := findStreamAmongCurrent(currentStreams, record)
	if found {
		*rec = currentStreams[ind]
		if currentStreams[ind].Finished {
			return found, true
		}
	}
	return found, false
}

// Create unique index `name` for collection of streams info
// If collection does not exist, DB will create it
func createIndexStreamInfoDB(db *Mongodb, db_name string) error {
	coll := db.client.Database(db_name).Collection(streams_info)

	_, err := coll.Indexes().CreateOne(
		context.Background(),
		mongo.IndexModel{
			Keys:    bson.D{{Key: "name", Value: 1}},
			Options: options.Index().SetUnique(true),
		},
	)
	return err
}

// Update streams_info collection with current StreamInfo
func updateStreamInfoDB(db *Mongodb, db_name string, rec StreamInfo) error {
	coll := db.client.Database(db_name).Collection(streams_info)

	_, err := coll.InsertOne(context.TODO(), rec)
	return err
}

// Fill stream information from different sources to the rec.
// Stream information is taken from collection streams_info
// If stream is not finished, only timestamp is valid
// If stream is not found in streams_info information is
// extracted form data collections
// If detailed is True, information is extracted from data collection
func updateStreamInfos(db *Mongodb, db_name string, rec *StreamsRecord, detailed bool) error {
	currentStreams, err := readStreamsInfo(db, db_name)
	if err != nil {
		return err
	}

	for i, record := range rec.Streams {
		found, finished := updateStreamInfoFromCurrent(currentStreams, record, &rec.Streams[i])
		if found && (!detailed || finished) { // Don't need to update if stream is finished
			continue
		}
		if !found { // set timestamp
			if err := fillInfoFromEarliestRecord(db, db_name, rec, record, i); err != nil {
				return err
			}
		}
		if detailed { // update last record (TimestampLast, stream finished flag)
			if err := fillInfoFromLastRecord(db, db_name, rec, record, i); err != nil {
				return err
			}
		}
		if !found || rec.Streams[i].Finished {
			// Inset stream info in DB if is it not yet there or stream is now finished
			err := updateStreamInfoDB(db, db_name, rec.Streams[i])
			// Error may happen if two brokers works with the same stream and try to
			// populate streams info.
			if err != nil {
				log.Debug("Update stream info in DB failed", err)
			}
		}
	}
	return nil
}

func sortRecords(rec *StreamsRecord) {
	sort.Slice(rec.Streams[:], func(i, j int) bool {
		return rec.Streams[i].Timestamp < rec.Streams[j].Timestamp
	})
}

func getStreamsInfoFromDb(db *Mongodb, db_name string, detailed bool) (StreamsRecord, error) {
	rec, err := readAllStreamNames(db, db_name)
	if err != nil {
		return StreamsRecord{}, err
	}

	err = updateStreamInfos(db, db_name, &rec, detailed)
	if err != nil {
		return StreamsRecord{}, err
	}

	sortRecords(&rec)
	return rec, nil
}

// Save StreamsRecord in global variable for caching purpose
func (ss *Streams) cacheStreams(db_name string, rec StreamsRecord) {
	if len(rec.Streams) > 0 {
		result := StreamsRecord{}
		utils.DeepCopy(rec, &result)
		ss.records[db_name] = result
		ss.lastUpdated[db_name] = time.Now()
	}
}

func filterStreamsByStatus(rec StreamInfo, streamStatus string) bool {
	return (rec.Finished && streamStatus == Stream_filter_finished) || (!rec.Finished && streamStatus == Stream_filter_unfinished)
}

func filterStreams(rec StreamsRecord, pars GetStreamsParams) []StreamInfo {

	if len(rec.Streams) == 0 {
		return rec.Streams
	}
	if pars.Last {
		return rec.Streams[len(rec.Streams)-1:]
	}
	limitedStreams := filterStreamsByName(rec, pars.From)
	if pars.Single {
		return limitedStreams[:1]
	}

	if pars.Filter == Stream_filter_all {
		return limitedStreams
	}
	nextStreams := limitedStreams[:0]
	for _, rec := range limitedStreams {
		if filterStreamsByStatus(rec, pars.Filter) {
			nextStreams = append(nextStreams, rec)
		}
	}
	return nextStreams
}

func filterStreamsByName(rec StreamsRecord, firstStream string) []StreamInfo {
	if firstStream != "" {
		ind := len(rec.Streams)
		for i, rec := range rec.Streams {
			if rec.Name == firstStream {
				ind = i
				break
			}
		}
		rec.Streams = rec.Streams[ind:]
	}
	return rec.Streams
}

func (ss *Streams) getStreams(db *Mongodb, request Request) (StreamsRecord, error) {
	params, ok := request.ExtraParam.(GetStreamsParams)
	if !ok {
		return StreamsRecord{}, errors.New(failed_extract_extra_params)
	}

	streamsLock.Lock()
	rec, err := ss.getStreamsInfoFromCache(request.DbName(), db.settings.UpdateStreamCachePeriodMs)
	streamsLock.Unlock()
	if err != nil {
		rec, err = getStreamsInfoFromDb(db, request.DbName(), params.Detailed)
	}
	if err != nil {
		return StreamsRecord{}, err
	}

	streamsLock.Lock()
	ss.cacheStreams(request.DbName(), rec)
	streamsLock.Unlock()

	rec.Streams = filterStreams(rec, params)
	return rec, nil
}
