//go:build integration_tests
// +build integration_tests

package database

import (
	"asapo_common/utils"
	"encoding/json"
	"fmt"
	"github.com/stretchr/testify/assert"
	"sync"
	"testing"
	"time"
)

type TestRecord struct {
	ID        int64             `bson:"message_id" json:"message_id"`
	Meta      map[string]string `bson:"meta" json:"meta"`
	Name      string            `bson:"name" json:"name"`
	Stream    string            `bson:"stream" json:"stream"`
	Substream int               `bson:"dataset_substream" json:"dataset_substream"`
	TimeID    int64             `bson:"time_id" json:"time_id"`
	Timestamp int64             `bson:"timestamp" json:"timestamp"`
}

type TestRecordStreamBtMeta struct {
	ID   string `bson:"_id" json:"_id"`
	Meta string `bson:"meta" json:"meta"`
}

var recbt = TestRecordStreamBtMeta{"bt", "meta_bt"}
var recst = TestRecordStreamBtMeta{"st_stream", "meta_st"}

type TestDataset struct {
	Stream    string       `bson:"stream" json:"stream"`
	Timestamp int64        `bson:"timestamp" json:"timestamp"`
	ID        int64        `bson:"message_id" json:"message_id"`
	Size      int64        `bson:"size" json:"size"`
	Messages  []TestRecord `bson:"messages" json:"messages"`
}

var db Mongodb

const beamtime = "bt"
const datasource = "12345"
const dbname = "bt_12345"
const stream_name = "stream"
const dataset_stream_name = "stream2"
const dbaddress = "127.0.0.1:27017"
const groupId = "bid2a5auidddp1vl71d0"
const metaID = "bt"

const badSymbolsDb = `/\."$`
const badSymbolsCol = `$`
const badSymbolsDbEncoded = "%2F%5C%2E%22%24"
const badSymbolsColEncoded = "%24"

var empty_next = map[string]string{"next_stream": ""}

var rec1 = TestRecord{1, empty_next, "aaa", stream_name, 1, 1, 0}
var rec1_later = TestRecord{1, empty_next, "aaa", stream_name, 1, 1, 1}
var rec_finished = TestRecord{2, map[string]string{"next_stream": "next1"}, finish_stream_keyword, stream_name, 1, 2, 2}
var rec2 = TestRecord{2, empty_next, "bbb", stream_name, 2, 2, 1}
var rec3 = TestRecord{3, empty_next, "ccc", stream_name, 3, 3, 2}
var rec_finished3 = TestRecord{3, map[string]string{"next_stream": "next1"}, finish_stream_keyword, stream_name, 3, 3, 2}
var rec_finished11 = TestRecord{11, map[string]string{"next_stream": "next1"}, finish_stream_keyword, stream_name, 3, 11, 2}

var rec1_expect = getExpectedRec(rec1)
var rec2_expect = getExpectedRec(rec2)
var rec3_expect = getExpectedRec(rec3)

var recs1 = SizeRecord{3}
var recs1_expect, _ = json.Marshal(recs1)
var recs2 = SizeRecord{0}
var recs2_expect, _ = json.Marshal(recs2)

// In order to kep back compatibility a `_id` is added to response
// function helps to match the `message_fallback` output
func getExpectedRec(rec TestRecord) []byte {
	var expected = map[string]interface{}{}
	expected["_id"] = rec.ID
	expected["message_id"] = rec.ID
	expected["meta"] = rec.Meta
	expected["name"] = rec.Name
	expected["stream"] = rec.Stream
	expected["dataset_substream"] = rec.Substream
	expected["time_id"] = rec.TimeID
	expected["timestamp"] = rec.Timestamp
	expected_rec, _ := json.Marshal(expected)
	return expected_rec
}

func getExtraParamNext(key string, resend bool, delay int, resendAttempts int) interface{} {
	var extraParamNext = ExtraParamNext{
		IdKey:          key,
		Resend:         resend,
		DelayMs:        delay,
		ResendAttempts: resendAttempts}
	return extraParamNext
}

func cleanup() {
	if db.client == nil {
		return
	}
	db.dropDatabase(dbname)
	db.Close()
}

func cleanupWithName(name string) {
	if db.client == nil {
		return
	}
	db.dropDatabase(name)
	db.Close()
}

// these are the integration tests. They assume mongo db is runnig on 127.0.0.1:27027
// test names should contain MongoDB*** so that go test could find them:
// go_integration_test(${TARGET_NAME}-connectdb "./..." "MongoDBConnect")
func TestMongoDBConnectFails(t *testing.T) {
	err := db.Connect("blabla")
	defer db.Close()
	assert.NotNil(t, err)
}

func TestMongoDBConnectOK(t *testing.T) {
	err := db.Connect(dbaddress)
	defer cleanup()
	assert.Nil(t, err)
}

func TestMongoDBGetNextErrorWhenNotConnected(t *testing.T) {
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next"})
	assert.Equal(t, utils.StatusServiceUnavailable, err.(*DBError).Code)
}

func TestMongoDBGetMetaErrorWhenNotConnected(t *testing.T) {
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "meta", ExtraParam: "0"})
	assert.Equal(t, utils.StatusServiceUnavailable, err.(*DBError).Code)
}

func TestMongoDBQueryMessagesErrorWhenNotConnected(t *testing.T) {
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "querymessages", ExtraParam: "0"})
	assert.Equal(t, utils.StatusServiceUnavailable, err.(*DBError).Code)
}

func TestMongoDBGetNextErrorWhenWrongDatabasename(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	_, err, _ := db.ProcessRequest(Request{Stream: stream_name, GroupId: groupId, Op: "next"})
	assert.Equal(t, utils.StatusWrongInput, err.(*DBError).Code)
}

func TestMongoDBGetNextErrorWhenNonExistingData(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource,
		Stream: "bla", GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":0,\"id_max\":0,\"next_stream\":\"\"}", err.Error())
}

func TestMongoDBGetLastErrorWhenNonExistingData(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: "bla", GroupId: groupId, Op: "last"})
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":0,\"id_max\":0,\"next_stream\":\"\"}", err.Error())
}

func TestMongoDBGetByIdErrorWhenNoData(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	extraParam := ExtraParamId{
		Id:    2,
		IdKey: "message_id"}
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "id", ExtraParam: extraParam})

	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":2,\"id_max\":0,\"next_stream\":\"\"}", err.Error())
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)

}

func TestMongoDBGetNextIfMissingID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec2)
	rec, _, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("time_id", false, 0, -1)})
	assert.Equal(t, string(rec2_expect), string(rec))

}

func TestMongoDBGetNextOK(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))
}

func TestMongoDBGetNextErrorOnFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.(*DBError).Message)
}

func TestMongoDBGetNextErrorOnFinishedStreamAlways(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.(*DBError).Message)
}

func TestMongoDBGetByIdErrorOnFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	extraParam := ExtraParamId{
		Id:    2,
		IdKey: "message_id"}
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "id", ExtraParam: extraParam})

	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.(*DBError).Message)
}

func TestMongoDBGetLastErrorOnFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "last"})
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.(*DBError).Message)
}

func TestMongoDBGetNextErrorOnNoMoreData(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"\"}", err.(*DBError).Message)
}

func TestMongoDBGetNextCorrectOrder(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec2)
	db.insertRecord(dbname, &rec1)
	res1, _, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	res2, _, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Equal(t, string(rec1_expect), string(res1))
	assert.Equal(t, string(rec2_expect), string(res2))
}

func getNOnes(array []int) int {
	sum1 := 0
	for _, result := range array {
		if result == 1 {
			sum1++
		}
	}
	return sum1
}

func insertRecords(n int) {
	records := make([]TestRecord, n)
	for ind, record := range records {
		record.ID = int64(ind) + 1
		record.TimeID = int64(ind) + 1
		record.Name = fmt.Sprint(ind)
		record.Stream = stream_name
		if err := db.insertRecord(dbname, &record); err != nil {
			fmt.Println("error at insert ", ind)
		}
	}
}

func getRecords(n int, resend bool) []int {
	results := make([]int, n)
	var wg sync.WaitGroup
	wg.Add(n)
	for i := 0; i < n; i++ {
		go func() {
			defer wg.Done()
			// fmt.Println("REquest ", string(encoded))
			res_bin, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
				GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
			if err != nil {
				fmt.Println("error at read ", i, string(res_bin))
			}
			var res TestRecord
			json.Unmarshal(res_bin, &res)
			if res.ID > 0 {
				results[res.ID-1] = 1
			}
		}()
	}
	wg.Wait()
	return results
}

func TestMongoDBGetNextInParallel(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	n := 100
	insertRecords(n)

	results := getRecords(n, false)

	assert.Equal(t, n, getNOnes(results))
}

func TestMongoDBGetNextInParallelWithResend(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 100})
	db.Connect(dbaddress)
	defer cleanup()
	n := 100
	insertRecords(n)

	results := getRecords(n, true)
	results2 := getRecords(n, true)

	assert.Equal(t, n, getNOnes(results), "first")
	assert.Equal(t, n, getNOnes(results2), "second")
}

func TestMongoDBGetLastAfterErasingDatabase(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	insertRecords(10)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	db.dropDatabase(dbname)

	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "last", ExtraParam: "0"})
	assert.Nil(t, err)
	assert.Equal(t, string(rec2_expect), string(res))
}

func TestMongoDBGetNextAfterErasingDatabase(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	insertRecords(200)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	db.dropDatabase(dbname)

	n := 100
	insertRecords(n)
	results := getRecords(n, false)
	assert.Equal(t, n, getNOnes(results))
}

func TestMongoDBGetNextEmptyAfterErasingDatabase(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	insertRecords(10)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	db.dropDatabase(dbname)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":0,\"id_max\":0,\"next_stream\":\"\"}", err.Error())
}

func TestMongoDBgetRecordByID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	extraParam := ExtraParamId{
		Id:    1,
		IdKey: "message_id"}
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "id", ExtraParam: extraParam})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))
}

func TestMongoDBgetRecordByIDFails(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	extraParam := ExtraParamId{
		Id:    2,
		IdKey: "message_id"}
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "id", ExtraParam: extraParam})
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":2,\"id_max\":1,\"next_stream\":\"\"}", err.Error())
}

func TestMongoDBGetRecordNext(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))
}

func TestMongoDBGetRecordNextMultipleStreams(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_dataset1)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	res_string, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", DatasetOp: true, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	var res_ds TestDataset
	json.Unmarshal(res_string, &res_ds)

	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))

	assert.Nil(t, err2)
	assert.Equal(t, rec_dataset1, res_ds)

}

func TestMongoDBGetRecordID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	extraParam := ExtraParamId{
		Id:    1,
		IdKey: "message_id"}
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "id", ExtraParam: extraParam})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))
}

func TestMongoDBWrongOp(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "bla"})
	assert.NotNil(t, err)
}

func TestMongoDBGetRecordLast(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "last", ExtraParam: "0"})
	assert.Nil(t, err)
	assert.Equal(t, string(rec2_expect), string(res))
}

func TestMongoDBGetNextAfterGetLastCorrect(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "last", ExtraParam: "0"})
	assert.Nil(t, err)
	assert.Equal(t, string(rec2_expect), string(res))

	db.insertRecord(dbname, &rec3)
	res, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))

}

func TestMongoDBGetGetLastInGroupCorrect(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)}) // to check it does not influence groupedlast

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))

	// first record - ok, then error
	res, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"\"}", err.(*DBError).Message)
	}
	// second record - ok, then error
	db.insertRecord(dbname, &rec2)
	res, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.Nil(t, err)
	assert.Equal(t, string(rec2_expect), string(res))
	res, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.NotNil(t, err)

	// stream finished - immediately error
	db.insertRecord(dbname, &rec_finished3)
	res, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":2,\"id_max\":2,\"next_stream\":\"next1\"}", err.(*DBError).Message)
	}

}

func TestMongoDBGetGetLastInGroupImmediateErrorOnFinishStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	db.insertRecord(dbname, &rec_finished3)
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.NotNil(t, err)
	_, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "groupedlast", ExtraParam: ""})
	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":2,\"id_max\":2,\"next_stream\":\"next1\"}", err.(*DBError).Message)
	}
}

func TestMongoDBGetSize(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	db.insertRecord(dbname, &rec3)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "size", ExtraParam: ExtraParamSize{false, false}})
	assert.Nil(t, err)
	assert.Equal(t, string(recs1_expect), string(res))
}

func TestMongoDBGetSizeWithFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "size", ExtraParam: ExtraParamSize{false, false}})
	assert.Nil(t, err)
	var rec_expect, _ = json.Marshal(&SizeRecord{1})
	assert.Equal(t, string(rec_expect), string(res))
}

func TestMongoDBGetSizeForDatasets(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "size", ExtraParam: ExtraParamSize{false, true}})
	assert.Equal(t, utils.StatusWrongInput, err.(*DBError).Code)

	_, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "size", ExtraParam: ExtraParamSize{true, true}})
	assert.Equal(t, utils.StatusWrongInput, err1.(*DBError).Code)
}

func TestMongoDBGetSizeForDatasetsWithFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_finished)

	res, _, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name, Op: "size", ExtraParam: ExtraParamSize{true, true}})

	var rec_expect, _ = json.Marshal(&SizeRecord{1})
	assert.Equal(t, string(rec_expect), string(res))
}

func TestMongoDBGetSizeDataset(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	db.insertRecord(dbname, &rec_dataset2_incomplete)

	size2_expect, _ := json.Marshal(SizeRecord{2})

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name, Op: "size", ExtraParam: ExtraParamSize{true, true}})
	assert.Nil(t, err)
	assert.Equal(t, string(size2_expect), string(res))
}

func TestMongoDBGetSizeNoRecords(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "size", ExtraParam: ExtraParamSize{false, false}})
	assert.Nil(t, err)
	assert.Equal(t, string(recs2_expect), string(res))
}

func TestMongoPing(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	err := db.Ping()
	assert.Nil(t, err)
}

func TestMongoPingNotConected(t *testing.T) {
	err := db.Ping()
	assert.NotNil(t, err)
}

func TestMongoDBgetRecordByIDNotConnected(t *testing.T) {
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "id", ExtraParam: "1"})
	assert.Equal(t, utils.StatusServiceUnavailable, err.(*DBError).Code)
}

func TestMongoDBResetCounter(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Nil(t, err1)
	assert.Equal(t, string(rec1_expect), string(res1))

	_, err_reset, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "resetcounter", ExtraParam: 1})
	assert.Nil(t, err_reset)

	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Nil(t, err2)
	assert.Equal(t, string(rec2_expect), string(res2))
}

func TestMongoDBGetMetaBtOK(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	rec_expect, _ := json.Marshal(recbt.Meta)
	db.insertMeta(dbname, &recbt)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: "whatever", Op: "meta", ExtraParam: "0"})

	assert.Nil(t, err)
	assert.Equal(t, string(rec_expect), string(res))
}

func TestMongoDBGetDatasets(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(beamtime+"_test01", &rec_dataset1)
	db.insertRecord(beamtime+"_test02", &rec_dataset1)
	db.insertRecord(beamtime+"_test03", &rec_dataset1)
	db.insertRecord("_test04", &rec_dataset1)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: "whatever", Op: "datasources", ExtraParam: "0"})

	assert.Nil(t, err)
	assert.Equal(t, "{\"sources\":[\"test01\",\"test02\",\"test03\"]}", string(res))

	db.dropDatabase(beamtime + "_test01")
	db.dropDatabase(beamtime + "_test02")
	db.dropDatabase(beamtime + "_test03")
}

func TestMongoDBGetMetaStOK(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	rec_expect, _ := json.Marshal(recst.Meta)
	db.insertMeta(dbname, &recst)

	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "meta", ExtraParam: "1"})

	assert.Nil(t, err)
	assert.Equal(t, string(rec_expect), string(res))
}

func TestMongoDBGetMetaErr(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "meta", ExtraParam: "1"})
	assert.NotNil(t, err)
	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
}

type MetaData struct {
	Temp    float64 `bson:"temp" json:"temp"`
	Counter int     `bson:"counter" json:"counter"`
	Text    string  `bson:"text" json:"text"`
}

type TestRecordMeta struct {
	ID     int      `bson:"message_id" json:"message_id"`
	FName  string   `bson:"fname" json:"fname"`
	Meta   MetaData `bson:"meta" json:"meta"`
	Stream string   `bson:"stream" json:"stream"`
}

var recq1 = TestRecordMeta{1, "aaa", MetaData{10.2, 10, "aaa"}, stream_name}
var recq2 = TestRecordMeta{2, "bbb", MetaData{11.2, 11, "bbb"}, stream_name}
var recq3 = TestRecordMeta{3, "ccc", MetaData{10.2, 10, "ccc"}, stream_name}
var recq4 = TestRecordMeta{4, "ddd", MetaData{13.2, 13, ""}, stream_name}

var tests = []struct {
	query string
	res   []TestRecordMeta
	ok    bool
}{
	{"message_id > 0", []TestRecordMeta{recq1, recq2, recq3, recq4}, true},
	{"meta.counter = 10", []TestRecordMeta{recq1, recq3}, true},
	{"meta.counter = 10 ORDER BY _id DESC", []TestRecordMeta{recq3, recq1}, true},
	{"meta.counter > 10 ORDER BY meta.counter DESC", []TestRecordMeta{recq4, recq2}, true},
	{"meta.counter = 18", []TestRecordMeta{}, true},
	{"meta.counter = 11", []TestRecordMeta{recq2}, true},
	{"meta.counter > 11", []TestRecordMeta{recq4}, true},
	{"meta.counter < 11", []TestRecordMeta{recq1, recq3}, true},
	{"meta.counter <= 11", []TestRecordMeta{recq1, recq2, recq3}, true},
	{"meta.counter >= 11", []TestRecordMeta{recq2, recq4}, true},
	{"meta.temp = 11.2", []TestRecordMeta{recq2}, true},
	{"meta.text = 'ccc'", []TestRecordMeta{recq3}, true},
	{"meta.text = ''", []TestRecordMeta{recq4}, true},
	{"meta.text = ccc", []TestRecordMeta{}, false},
	{"meta.text != 'ccc'", []TestRecordMeta{recq1, recq2, recq4}, true},
	{"meta.temp BETWEEN 11 AND 13", []TestRecordMeta{recq2}, true},
	{"meta.temp not BETWEEN 11 and 13", []TestRecordMeta{recq1, recq3, recq4}, true},
	{"meta.counter IN (10,13)", []TestRecordMeta{recq1, recq3, recq4}, true},
	{"meta.counter NOT IN (10,13)", []TestRecordMeta{recq2}, true},
	{"meta.text IN ('aaa','ccc')", []TestRecordMeta{recq1, recq3}, true},
	{"message_id = 1", []TestRecordMeta{recq1}, true},
	{"meta.text REGEXP '^c+'", []TestRecordMeta{recq3}, true},
	{"meta.text REGEXP '^a|b'", []TestRecordMeta{recq1, recq2}, true},
	// mongo 4.07+ is needed for NOT REXEXP
	{"meta.text NOT REGEXP '^c+'", []TestRecordMeta{recq1, recq2, recq4}, true},
	{"give_error", []TestRecordMeta{}, false},
	{"meta.counter = 10 AND meta.text = 'ccc'", []TestRecordMeta{recq3}, true},
	{"meta.counter = 10 OR meta.text = 'bbb'", []TestRecordMeta{recq1, recq2, recq3}, true},
	{"(meta.counter = 10 OR meta.counter = 11) AND (meta.text = 'bbb' OR meta.text = 'ccc')", []TestRecordMeta{recq2, recq3}, true},
	{"(meta.counter = 10 OR meta.counter = 11 AND (meta.text = 'bbb' OR meta.text = 'ccc')", []TestRecordMeta{}, false},
}

func TestMongoDBQueryMessagesOK(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	//	logger.SetLevel(logger.DebugLevel)
	db.insertRecord(dbname, &recq1)
	db.insertRecord(dbname, &recq2)
	db.insertRecord(dbname, &recq3)
	db.insertRecord(dbname, &recq4)

	for _, test := range tests {
		//		info, _ := db.client.BuildInfo()
		//		if strings.Contains(test.query, "NOT REGEXP") && !info.VersionAtLeast(4, 0, 7) {
		//			fmt.Println("Skipping NOT REGEXP test since it is not supported by this mongodb version")
		//			continue
		//		}

		res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "querymessages", ExtraParam: test.query})
		var res []TestRecordMeta
		json.Unmarshal(res_string, &res)
		//		fmt.Println(string(res_string))
		if test.ok {
			assert.Nil(t, err, test.query)
			assert.Equal(t, test.res, res)
		} else {
			assert.NotNil(t, err, test.query)
			assert.Equal(t, 0, len(res))
		}
	}

}

func TestMongoDBQueryMessagesOnEmptyDatabase(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	for _, test := range tests {
		res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, Op: "querymessages", ExtraParam: test.query})
		var res []TestRecordMeta
		json.Unmarshal(res_string, &res)
		assert.Equal(t, 0, len(res))
		if test.ok {
			assert.Nil(t, err, test.query)
		} else {
			assert.NotNil(t, err, test.query)
		}
	}
}

var rec_dataset1 = TestDataset{dataset_stream_name, 0, 1, 3, []TestRecord{rec1, rec2, rec3}}
var rec_dataset1_incomplete = TestDataset{dataset_stream_name, 1, 1, 4, []TestRecord{rec1, rec2, rec3}}
var rec_dataset2_incomplete = TestDataset{dataset_stream_name, 2, 2, 4, []TestRecord{rec1, rec2, rec3}}
var rec_dataset2 = TestDataset{dataset_stream_name, 1, 2, 4, []TestRecord{rec1, rec2, rec3}}
var rec_dataset3 = TestDataset{dataset_stream_name, 2, 3, 3, []TestRecord{rec3, rec2, rec2}}

var rec_dataset2_incomplete3 = TestDataset{stream_name, 1, 2, 3, []TestRecord{rec1, rec2}}

func TestMongoDBGetDataset(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", DatasetOp: true, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset1, res)
}

func TestMongoDBNoDataOnNotCompletedFirstDataset(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1_incomplete)
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name, GroupId: groupId, Op: "next",
		DatasetOp: true, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Equal(t, utils.StatusPartialData, err.(*DBError).Code)
	var res TestDataset
	json.Unmarshal([]byte(err.(*DBError).Message), &res)
	assert.Equal(t, rec_dataset1_incomplete, res)
}

func TestMongoDBNoDataOnNotCompletedNextDataset(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_dataset2_incomplete)
	_, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", DatasetOp: true, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	_, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", DatasetOp: true, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})

	assert.Equal(t, utils.StatusPartialData, err1.(*DBError).Code)
	assert.Equal(t, utils.StatusPartialData, err2.(*DBError).Code)
	var res TestDataset
	json.Unmarshal([]byte(err2.(*DBError).Message), &res)
	assert.Equal(t, rec_dataset2_incomplete, res)
}

func convertToDataset(dataset_str []byte) TestDataset {
	var res TestDataset
	json.Unmarshal(dataset_str, &res)
	return res
}

func TestMongoDBNextDatasetPartial(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_dataset2_incomplete)
	db.insertRecord(dbname, &rec_dataset3)

	res_string, _, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "id", DatasetOp: true, MinDatasetSize: 1, Substream: 1, ExtraParam: ExtraParamId{Id: 2, IdKey: "message_id"}})
	res1 := convertToDataset(res_string)

	res_string, _, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", DatasetOp: true, MinDatasetSize: 1, Substream: 2, ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	res2 := convertToDataset(res_string)

	res_string, _, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "last", DatasetOp: true, MinDatasetSize: 1, Substream: 2, ExtraParam: nil})
	res3 := convertToDataset(res_string)

	var rec1_dataset_expected = TestDataset{dataset_stream_name, 2, 2, 4, []TestRecord{rec1}}
	var rec2_dataset_expected = TestDataset{dataset_stream_name, 1, 1, 4, []TestRecord{rec2}}
	var rec3_dataset_expected = TestDataset{dataset_stream_name, 2, 3, 3, []TestRecord{rec2}}

	assert.Equal(t, rec1_dataset_expected, res1)
	assert.Equal(t, rec2_dataset_expected, res2)
	assert.Equal(t, rec3_dataset_expected, res3)
}

func TestMongoDBGetRecordLastDataSetSkipsIncompleteSets(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	db.insertRecord(dbname, &rec_dataset2)

	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "last", DatasetOp: true, ExtraParam: "0"})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset1, res)
}

func TestMongoDBGetRecordLastDataSetReturnsIncompleteSets(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_dataset2_incomplete)

	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "last",
		DatasetOp: true, MinDatasetSize: 3, ExtraParam: "0"})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset2_incomplete, res)
}

func TestMongoDBGetRecordLastDataSetSkipsIncompleteSetsWithMinSize(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	db.insertRecord(dbname, &rec_dataset2_incomplete3)

	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "last",
		DatasetOp: true, MinDatasetSize: 3, ExtraParam: "0"})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)
	assert.Equal(t, rec_dataset1, res)
}

func TestMongoDBGetRecordLastDataSetWithFinishedStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	db.insertRecord(dbname, &rec_finished)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "last",
		DatasetOp: true, ExtraParam: "0"})

	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.Error())
	}
}

func TestMongoDBGetRecordLastDataSetWithIncompleteDatasetsAndFinishedStreamReturnsEndofStream(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_finished)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "last",
		DatasetOp: true, MinDatasetSize: 2, ExtraParam: "0"})

	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"next1\"}", err.Error())
	}
}

func TestMongoDBGetRecordLastDataSetOK(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec_dataset1)
	db.insertRecord(dbname, &rec_dataset3)

	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "last", DatasetOp: true, ExtraParam: "0"})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset3, res)
}

func TestMongoDBGetDatasetID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec_dataset1)
	extraParam := ExtraParamId{
		Id:    1,
		IdKey: "message_id"}
	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "id", DatasetOp: true, ExtraParam: extraParam})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset1, res)

}

func TestMongoDBErrorOnIncompleteDatasetID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec_dataset1_incomplete)
	extraParam := ExtraParamId{
		Id:    1,
		IdKey: "message_id"}
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "id", DatasetOp: true, ExtraParam: extraParam})

	assert.Equal(t, utils.StatusPartialData, err.(*DBError).Code)

	var res TestDataset
	json.Unmarshal([]byte(err.(*DBError).Message), &res)

	assert.Equal(t, rec_dataset1_incomplete, res)

}

func TestMongoDBGetNextRecoverAfterIncompleteDataset(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec_dataset1_incomplete)
	db.insertRecord(dbname, &rec_dataset2_incomplete)

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3),
		DatasetOp: true, MinDatasetSize: 0})
	assert.Equal(t, utils.StatusPartialData, err.(*DBError).Code)

	db.replaceRecord(dbname, 1, &rec_dataset1)
	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3),
		DatasetOp: true, MinDatasetSize: 0})

	assert.Nil(t, err)

	var res_ds TestDataset
	json.Unmarshal(res_string, &res_ds)
	assert.Equal(t, rec_dataset1, res_ds)
}

func TestMongoDBGetNextUserKeyMissingID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec2)
	db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})

	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
	assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":2,\"next_stream\":\"\"}", err.Error())

	db.insertRecord(dbname, &rec1)
	rec1, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	rec2, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	assert.Equal(t, string(rec2_expect), string(rec1))
	assert.Equal(t, string(rec1_expect), string(rec2))
}

func TestMongoDBOkOnIncompleteDatasetID(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec_dataset1_incomplete)
	extraParam := ExtraParamId{
		Id:    1,
		IdKey: "message_id"}
	res_string, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: dataset_stream_name,
		GroupId: groupId, Op: "id", DatasetOp: true, MinDatasetSize: 3, ExtraParam: extraParam})

	assert.Nil(t, err)

	var res TestDataset
	json.Unmarshal(res_string, &res)

	assert.Equal(t, rec_dataset1_incomplete, res)

}

type Stream struct {
	name    string
	records []TestRecord
}

var testsStreams = []struct {
	from            string
	streams         []Stream
	expectedStreams StreamsRecord
	test            string
	ok              bool
}{
	{"", []Stream{}, StreamsRecord{[]StreamInfo{}}, "no streams", true},
	{"", []Stream{{"ss1", []TestRecord{rec2, rec1}}},
		StreamsRecord{[]StreamInfo{StreamInfo{Name: "ss1", Timestamp: 0, LastId: 2, TimestampLast: 1}}}, "one stream", true},
	{"", []Stream{{"ss1", []TestRecord{rec2, rec1}},
		{"ss2", []TestRecord{rec2, rec3}}},
		StreamsRecord{[]StreamInfo{StreamInfo{Name: "ss1", Timestamp: 0, LastId: 2, TimestampLast: 1},
			StreamInfo{Name: "ss2", Timestamp: 1, LastId: 3, TimestampLast: 2}}}, "two streams", true},
	{"ss2", []Stream{{"ss1", []TestRecord{rec1, rec2}}, {"ss2", []TestRecord{rec2, rec3}}}, StreamsRecord{[]StreamInfo{StreamInfo{Name: "ss2", Timestamp: 1, LastId: 3, TimestampLast: 2}}}, "with from", true},
	{"", []Stream{{"ss1$", []TestRecord{rec2, rec1}}},
		StreamsRecord{[]StreamInfo{StreamInfo{Name: "ss1$", Timestamp: 0, LastId: 2, TimestampLast: 1}}}, "one stream encoded", true},
	{"ss2$", []Stream{{"ss1$", []TestRecord{rec1, rec2}}, {"ss2$", []TestRecord{rec2, rec3}}}, StreamsRecord{[]StreamInfo{StreamInfo{Name: "ss2$", Timestamp: 1, LastId: 3, TimestampLast: 2}}}, "with from encoded", true},
}

func TestMongoDBListStreams(t *testing.T) {
	for _, test := range testsStreams {
		db.Connect(dbaddress)
		for _, stream := range test.streams {
			for _, rec := range stream.records {
				rec.Stream = stream.name
				db.insertRecord(dbname, &rec)
			}
		}
		var rec_streams_expect, _ = json.Marshal(test.expectedStreams)

		res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: "0", Op: "streams",
			ExtraParam: GetStreamsParams{From: test.from, Filter: "all", Detailed: true, Last: false, Single: false}})
		if test.ok {
			assert.Nil(t, err, test.test)
			assert.Equal(t, string(rec_streams_expect), string(res), test.test)
		} else {
			assert.NotNil(t, err, test.test)
		}
		cleanup()
	}
}

func TestMongoDBAckMessage(t *testing.T) {
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)

	request := Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "ackmessage", ExtraParam: ExtraParamAck{Id: 1}}
	res, err, _ := db.ProcessRequest(request)
	nacks, _ := db.getNacks(request, 0, 0)
	assert.Nil(t, err)
	assert.Equal(t, "", string(res))
	assert.Equal(t, 0, len(nacks))
}

var testsNacs = []struct {
	params        ExtraParamNacks
	resString     string
	insertRecords bool
	ackRecords    bool
	ok            bool
	test          string
}{
	{ExtraParamNacks{From: 1, To: 0}, "{\"unacknowledged\":[1,2,3,4,5,6,7,8,9,10]}", true, false, true, "whole range"},
	{ExtraParamNacks{From: 1, To: 5}, "{\"unacknowledged\":[1,2,3,4,5]}", true, false, true, "to given"},
	{ExtraParamNacks{From: 5, To: 0}, "{\"unacknowledged\":[5,6,7,8,9,10]}", true, false, true, "from given"},
	{ExtraParamNacks{From: 3, To: 7}, "{\"unacknowledged\":[3,4,5,6,7]}", true, false, true, "range given"},
	{ExtraParamNacks{From: 1, To: 1}, "{\"unacknowledged\":[1]}", true, false, true, "single record"},
	{ExtraParamNacks{From: 3, To: 1}, "{\"unacknowledged\":[]}", true, false, false, "to lt from"},
	{ExtraParamNacks{From: 1, To: 0}, "{\"unacknowledged\":[]}", false, false, true, "no records"},
	{ExtraParamNacks{From: 1, To: 0}, "{\"unacknowledged\":[1,5,6,7,8,9,10]}", true, true, true, "skip acks"},
	{ExtraParamNacks{From: 2, To: 4}, "{\"unacknowledged\":[]}", true, true, true, "all acknowledged"},
	{ExtraParamNacks{From: 1, To: 4}, "{\"unacknowledged\":[1]}", true, true, true, "some acknowledged"},
}

func TestMongoDBNacks(t *testing.T) {
	for _, test := range testsNacs {
		db.Connect(dbaddress)
		if test.insertRecords {
			insertRecords(10)
			db.insertRecord(dbname, &rec_finished11)
		}
		if test.ackRecords {
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 2}})
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 3}})
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 4}})
		}

		res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "nacks", ExtraParam: test.params})
		if test.ok {
			assert.Nil(t, err, test.test)
			assert.Equal(t, test.resString, string(res), test.test)
		} else {
			assert.NotNil(t, err, test.test)
		}
		cleanup()
	}
}

var testsLastAcs = []struct {
	insertRecords bool
	ackRecords    bool
	resString     string
	test          string
}{
	{false, false, "{\"lastAckId\":0}", "empty db"},
	{true, false, "{\"lastAckId\":0}", "no acks"},
	{true, true, "{\"lastAckId\":4}", "last ack 4"},
}

func TestMongoDBLastAcks(t *testing.T) {
	for _, test := range testsLastAcs {
		db.Connect(dbaddress)
		db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})
		if test.insertRecords {
			insertRecords(10)
			db.insertRecord(dbname, &rec_finished11)
		}
		if test.ackRecords {
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 2}})
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 3}})
			db.ackRecord(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, ExtraParam: ExtraParamAck{Id: 4}})
		}

		res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "lastack"})
		assert.Nil(t, err, test.test)
		assert.Equal(t, test.resString, string(res), test.test)
		cleanup()
	}
}

func TestMongoDBGetNextUsesInprocessedImmedeatly(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Nil(t, err2)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec1_expect), string(res1))
	assert.Equal(t, string(rec2_expect), string(res2))
}

func TestMongoDBGetNextUsesInprocessedNumRetry(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	_, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.NotNil(t, err2)
	if err2 != nil {
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"\"}", err2.Error())
	}
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec1_expect), string(res1))
}

func TestMongoDBGetNextUsesInprocessedAfterTimeout(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	time.Sleep(time.Second * 3)
	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Nil(t, err2)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec2_expect), string(res1))
	assert.Equal(t, string(rec1_expect), string(res2))
}

func TestMongoDBGetNextReturnsToNormalAfterUsesInprocessed(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	db.insertRecord(dbname, &rec_finished3)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	time.Sleep(time.Second)
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 1000, 3)})
	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Nil(t, err2)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec1_expect), string(res1))
	assert.Equal(t, string(rec2_expect), string(res2))
}

func TestMongoDBGetNextUsesInprocessedImmedeatlyIfFinishedStream(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 10})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec_finished)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec1_expect), string(res1))
}

func TestMongoDBGetNextUsesInprocessedImmedeatlyIfEndofStream(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 10})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Nil(t, err2)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec2_expect), string(res1))
	assert.Equal(t, string(rec1_expect), string(res2))
}

func TestMongoDBAckDeletesInprocessed(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})

	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "ackmessage", ExtraParam: ExtraParamAck{Id: 1}})
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name,
		GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 3)})
	assert.NotNil(t, err)
	if err != nil {
		assert.Equal(t, utils.StatusNoData, err.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"\"}", err.Error())
	}
}

func TestMongoDBAckTwiceErrors(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	db.insertRecord(dbname, &rec1)
	query_str := "{\"Id\":1,\"Op\":\"ackmessage\"}"
	db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "ackmessage", ExtraParam: query_str})
	_, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "ackmessage", ExtraParam: query_str})
	assert.Equal(t, utils.StatusWrongInput, err.(*DBError).Code)
}

func TestMongoDBNegAck(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()

	db.insertRecord(dbname, &rec1)
	params := ExtraParamNegAck{Id: 1, DelayMs: 0}
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "negackmessage", ExtraParam: params})
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)}) // first time message from negack
	_, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})  // second time nothing
	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "negackmessage", ExtraParam: params})
	_, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)}) // second time nothing

	assert.Nil(t, err)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.NotNil(t, err1)
	assert.Nil(t, err2)

	if err1 != nil {
		assert.Equal(t, utils.StatusNoData, err1.(*DBError).Code)
		assert.Equal(t, "{\"op\":\"get_record_by_id\",\"id\":1,\"id_max\":1,\"next_stream\":\"\"}", err1.Error())
	}
}

func TestMongoDBGetNextClearsInprocessAfterReset(t *testing.T) {
	db.SetSettings(DBSettings{ReadFromInprocessPeriod: 0})
	db.Connect(dbaddress)
	defer cleanup()
	err := db.insertRecord(dbname, &rec1)
	db.insertRecord(dbname, &rec2)
	res, err, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	res1, err1, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "resetcounter", ExtraParam: "0"})
	res2, err2, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})
	res3, err3, _ := db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: stream_name, GroupId: groupId, Op: "next", ExtraParam: getExtraParamNext("message_id", true, 0, 1)})

	assert.Nil(t, err)
	assert.Nil(t, err1)
	assert.Nil(t, err2)
	assert.Nil(t, err3)
	assert.Equal(t, string(rec1_expect), string(res))
	assert.Equal(t, string(rec1_expect), string(res1))
	assert.Equal(t, string(rec1_expect), string(res2))
	assert.Equal(t, string(rec1_expect), string(res3))
}

var testsDeleteStream = []struct {
	stream  string
	params  ExtraParamDelete
	ok      bool
	ok2     bool
	message string
}{
	{"test", ExtraParamDelete{true, true}, true, false, "delete stream"},
	{"test", ExtraParamDelete{false, true}, true, true, "delete stream"},
	{`test$/\  .%&?*#'`, ExtraParamDelete{false, true}, true, true, "delete stream"},
}

func TestDeleteStreams(t *testing.T) {
	defer cleanup()
	for _, test := range testsDeleteStream {
		db.Connect(dbaddress)
		test_rec := rec1
		test_rec.Stream = test.stream
		db.insertRecord(dbname, &test_rec)
		db.createIndices(Request{Beamtime: beamtime, DataSource: datasource})
		db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: test.stream, GroupId: "123", Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
		request := Request{Beamtime: beamtime, DataSource: datasource, Stream: test.stream, GroupId: groupId, Op: "ackmessage", ExtraParam: ExtraParamAck{Id: 1}}
		_, err, _ := db.ProcessRequest(request)
		assert.Nil(t, err, test.message)
		_, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: test.stream, GroupId: "", Op: "delete_stream", ExtraParam: test.params})
		_, err, _ = db.ProcessRequest(Request{Beamtime: beamtime, DataSource: datasource, Stream: test.stream, GroupId: "", Op: "delete_stream", ExtraParam: test.params})
		if test.ok {
			pars := GetStreamsParams{From: "", Filter: "all", Detailed: true, Last: false, Single: false}
			rec, err := streams.getStreams(&db, Request{Beamtime: beamtime, DataSource: datasource, ExtraParam: pars})
			assert.Equal(t, 0, len(rec.Streams), test.message)
			assert.Nil(t, err, test.message)
		}
		if test.ok2 {
			assert.Nil(t, err, test.message+" 2")
		} else {
			assert.Equal(t, utils.StatusWrongInput, err.(*DBError).Code, test.message+" 2")
		}
	}
}

var testsEncodings = []struct {
	beamtime    string
	datasource  string
	stream      string
	group       string
	dbname_indb string
	stream_indb string
	group_indb  string
	message     string
	ok          bool
}{
	{"bt", "dbname", "col", "group", "bt_dbname", "col", "group", "no encoding", true},
	{"bt", "dbname" + badSymbolsDb, "col", "group", "bt_dbname" + badSymbolsDbEncoded, "col", "group", "symbols in db", true},
	{"bt", "dbname", "col" + badSymbolsCol, "group" + badSymbolsCol, "bt_dbname", "col" + badSymbolsColEncoded, "group" + badSymbolsColEncoded, "symbols in col", true},
	{"bt", "dbname" + badSymbolsDb, "col" + badSymbolsCol, "group" + badSymbolsCol, "bt_dbname" + badSymbolsDbEncoded, "col" + badSymbolsColEncoded, "group" + badSymbolsColEncoded, "symbols in col and db", true},
}

func TestMongoDBEncodingOK(t *testing.T) {
	for _, test := range testsEncodings {
		db.Connect(dbaddress)
		rec_ := rec1
		rec_.Stream = test.stream
		db.insertRecord(test.dbname_indb, &rec_)
		res, err, _ := db.ProcessRequest(Request{Beamtime: test.beamtime, DataSource: test.datasource, Stream: test.stream,
			GroupId: test.group, Op: "next", ExtraParam: getExtraParamNext("message_id", false, 0, -1)})
		if test.ok {
			assert.Nil(t, err, test.message)
			rec_json := getExpectedRec(rec_)
			assert.Equal(t, string(rec_json), string(res), test.message)
		} else {
			assert.Equal(t, utils.StatusWrongInput, err.(*DBError).Code, test.message)
		}
		cleanupWithName(test.dbname_indb)
	}
}
