package server

import (
	"asapo_broker/database"
	"asapo_common/logger"
	"github.com/stretchr/testify/mock"
	"github.com/stretchr/testify/suite"
	"net/http"
	"net/url"
	"strings"
	"testing"
)

type GetCommandsTestSuite struct {
	suite.Suite
	mock_db *database.MockedDatabase
}

func (suite *GetCommandsTestSuite) SetupTest() {
	statistics.Reset()
	suite.mock_db = new(database.MockedDatabase)
	db = suite.mock_db
	prepareTestAuth()
	logger.SetMockLog()
}

func (suite *GetCommandsTestSuite) TearDownTest() {
	assertExpectations(suite.T(), suite.mock_db)
	logger.UnsetMockLog()
	db = nil
}

func TestGetCommandsTestSuite(t *testing.T) {
	suite.Run(t, new(GetCommandsTestSuite))
}

var testsGetCommand = []struct {
	command       string
	source        string
	stream        string
	groupid       string
	reqString     string
	queryParams   string
	externalParam interface{}
}{
	{"last", expectedSource, expectedStream, "", expectedStream + "/0/last", "", nil},
	{"id", expectedSource, expectedStream, "", expectedStream + "/0/1", "&id=1", database.ExtraParamId{Id: 1, IdKey: "message_id"}},
	{"meta", expectedSource, "default", "", "default/0/meta/0", "", "0"},
	{"nacks", expectedSource, expectedStream, expectedGroupID, expectedStream + "/" + expectedGroupID + "/nacks", "", database.ExtraParamNacks{From: 1, To: 0}},
	{"groupedlast", expectedSource, expectedStream, expectedGroupID, expectedStream + "/" + expectedGroupID + "/groupedlast", "", nil},
	{"next", expectedSource, expectedStream, expectedGroupID, expectedStream + "/" + expectedGroupID + "/next", "", database.ExtraParamNext{IdKey: "message_id", Resend: false, DelayMs: 0, ResendAttempts: -1}},
	{"next", expectedSource, expectedStream, expectedGroupID, expectedStream + "/" +
		expectedGroupID + "/next", "&resend_nacks=true&delay_ms=10000&resend_attempts=3", database.ExtraParamNext{IdKey: "message_id", Resend: true, DelayMs: 10000, ResendAttempts: 3}},
	{"size", expectedSource, expectedStream, "", expectedStream + "/size", "", database.ExtraParamSize{false, false}},
	{"size", expectedSource, expectedStream, "", expectedStream + "/size", "&incomplete=true", database.ExtraParamSize{true, true}},
	{"streams", expectedSource, "0", "", "0/streams", "", database.GetStreamsParams{From: "", Filter: "all", Detailed: true, Last: false, Single: false}},
	{"lastack", expectedSource, expectedStream, expectedGroupID, expectedStream + "/" + expectedGroupID + "/lastack", "", nil},
	{"datasources", expectedSource, "0", "", "0/sources", "", nil},
}

func (suite *GetCommandsTestSuite) TestGetCommandsCallsCorrectRoutine() {
	for _, test := range testsGetCommand {
		suite.mock_db.On("ProcessRequest", database.Request{Beamtime: expectedBeamtimeId, DataSource: test.source, Stream: test.stream, GroupId: test.groupid, Op: test.command, ExtraParam: test.externalParam}).Return([]byte("Hello"), nil)
		logger.MockLog.On("WithFields", mock.MatchedBy(containsMatcherMap(test.command)))
		logger.MockLog.On("Debug", mock.Anything)

		w := doRequest("/beamtime/" + expectedBeamtimeId + "/" + test.source + "/" + test.reqString + correctTokenSuffix + test.queryParams)
		suite.Equal(http.StatusOK, w.Code, test.command+" OK")
		suite.Equal("Hello", string(w.Body.Bytes()), test.command+" sends data")
	}
}

func (suite *GetCommandsTestSuite) TestGetCommandsCorrectlyProcessedEncoding() {
	badSymbols := "%$&./\\_$&\""
	for _, test := range testsGetCommand {
		newstream := test.stream + badSymbols
		newsource := test.source + badSymbols
		newgroup := ""
		if test.groupid != "" {
			newgroup = test.groupid + badSymbols
		}
		encodedStream := url.PathEscape(newstream)
		encodedSource := url.PathEscape(newsource)
		encodedGroup := url.PathEscape(newgroup)
		test.reqString = strings.Replace(test.reqString, test.groupid, encodedGroup, 1)
		test.reqString = strings.Replace(test.reqString, test.source, encodedSource, 1)
		test.reqString = strings.Replace(test.reqString, test.stream, encodedStream, 1)
		suite.mock_db.On("ProcessRequest", database.Request{Beamtime: expectedBeamtimeId, DataSource: newsource, Stream: newstream, GroupId: newgroup, Op: test.command, ExtraParam: test.externalParam}).Return([]byte("Hello"), nil)
		logger.MockLog.On("WithFields", mock.MatchedBy(containsMatcherMap(test.command)))
		logger.MockLog.On("Debug", mock.MatchedBy(containsMatcherStr("got request")))
		w := doRequest("/beamtime/" + expectedBeamtimeId + "/" + encodedSource + "/" + test.reqString + correctTokenSuffix + test.queryParams)
		suite.Equal(http.StatusOK, w.Code, test.command+" OK")
		suite.Equal("Hello", string(w.Body.Bytes()), test.command+" sends data")
	}
}
