//go:build integration_tests
// +build integration_tests

package server

import (
	"asapo_broker/database"
	"asapo_common/utils"
	"encoding/json"
	"fmt"
	"github.com/stretchr/testify/assert"
	"net/http"
	"os"
	"testing"
)

const beamtime = "bt"
const datasource = "12345"
const stream_name = "stream"
const dbaddress = "127.0.0.1:27017"
const groupId = "bid2a5auidddp1vl71d0"
const dbname = beamtime + "_" + datasource
const base_path = "/beamtime/" + beamtime + "/" + datasource + "/" + stream_name + "/"

func cleanup() {
	db.DropDatabase(dbname)
}

func TestMain(m *testing.M) {
	// Setup: initialize the server/router
	InitDB(new(database.Mongodb))
	db.Connect(dbaddress)

	// Run the tests
	code := m.Run()

	// Teardown: perform any necessary cleanup (if required)
	db.DropDatabase(dbname)
	db.Close()

	os.Exit(code)
}

func toJsonString(msg interface{}) string {
	var body, _ = json.Marshal(msg)
	return string(body)
}

type DatasetMessage struct {
	Id        int64                 `bson:"message_id" json:"_id"`
	Size      int32                 `bson:"size" json:"size"`
	TimeId    int64                 `bson:"time_id" json:"time_id"`
	Stream    string                `bson:"stream" json:"stream"`
	Timestamp int64                 `bson:"timestamp" json:"timestamp"`
	Messages  []database.MessageRaw `bson:"messages" json:"messages"`
}

func TestServerNextDatasetPartial(t *testing.T) {
	defer cleanup()
	for idx := int64(1); idx < 3; idx++ {
		for sub_idx := 1; sub_idx < 3; sub_idx++ {
			doRequest(base_path+"write_message?token=token&id="+fmt.Sprint(idx)+"&dataset_size=4", "POST",
				toJsonString(database.MessageRaw{Id: idx, DatasetSubstream: sub_idx, Size: 4, Stream: stream_name}))
		}
	}

	rr := doRequest(base_path+groupId+"/next?token=token&dataset=true&minsize=1&substream=2", "GET")

	var message DatasetMessage
	if err := json.Unmarshal(rr.Body.Bytes(), &message); err != nil {
		fmt.Println("failed to deserialize request body: " + err.Error())
	}

	assert.Equal(t, http.StatusOK, rr.Code)
	assert.Equal(t, int64(1), message.Id)
	assert.Equal(t, 2, message.Messages[0].DatasetSubstream)

}

func TestServerInsertMessageWrongSubstream(t *testing.T) {
	defer cleanup()
	res := doRequest(base_path+"write_message?token=token&id=1", "POST",
		toJsonString(database.MessageRaw{Id: 1, DatasetSubstream: 2, Size: 0, Stream: stream_name}))
	assert.Equal(t, utils.StatusWrongInput, res.Code)
	assert.Equal(t, "DatasetSubstream > 0", res.Body.String())
}

func TestServerInsertDatasetDuplicatedDatasetId(t *testing.T) {
	defer cleanup()
	res := doRequest(base_path+"write_message?token=token&id=1&dataset_size=4", "POST",
		toJsonString(database.MessageRaw{Id: 1, DatasetSubstream: 2, Size: 4, Stream: stream_name}))
	assert.Equal(t, http.StatusOK, res.Code)

	res = doRequest(base_path+"write_message?token=token&id=1&dataset_size=4", "POST",
		toJsonString(database.MessageRaw{Id: 1, DatasetSubstream: 2, Size: 4, Stream: stream_name}))
	assert.Equal(t, utils.StatusWrongInput, res.Code)
}

func TestServerInsertDatasetDuplicatedMessageId(t *testing.T) {
	defer cleanup()
	res := doRequest(base_path+"write_message?token=token&id=1", "POST",
		toJsonString(database.MessageRaw{Id: 1, DatasetSubstream: 0, Size: 0, Stream: stream_name}))
	assert.Equal(t, http.StatusOK, res.Code)

	res = doRequest(base_path+"write_message?token=token&id=1&dataset_size=4", "POST",
		toJsonString(database.MessageRaw{Id: 1, DatasetSubstream: 2, Size: 4, Stream: stream_name}))
	assert.Equal(t, utils.StatusWrongInput, res.Code)
}
