#ifndef ASAPO_CONSUMER_IMPL_H
#define ASAPO_CONSUMER_IMPL_H

#include "asapo/common/networking.h"
#include <mutex>
#include <atomic>
#include "asapo/consumer/consumer.h"
#include "asapo/io/io.h"
#include "asapo/http_client/http_client.h"
#include "asapo/http_request/service_request.h"
#include "net_client.h"

namespace asapo {

enum class GetMessageServerOperation {
    GetNext,
    GetNextAvailable,
    GetLast,
    GetID,
    GetLastInGroup,
};

DataSet DecodeDatasetFromResponse(std::string response, Error* err);
class ConsumerImpl final : public asapo::Consumer {
  public:
    explicit ConsumerImpl(std::string server_uri, std::string source_path, bool has_filesystem,
                          SourceCredentials source);

    Error Acknowledge(std::string group_id, uint64_t id, std::string) override;
    Error NegativeAcknowledge(std::string group_id, uint64_t id, uint64_t delay_ms,
                              std::string stream) override;

    IdList GetUnacknowledgedMessages(std::string group_id,
                                     uint64_t from_id,
                                     uint64_t to_id,
                                     std::string stream,
                                     Error* error) override;

    uint64_t GetLastAcknowledgedMessage(std::string group_id, std::string stream, Error* error) override;

    Error ResetLastReadMarker(std::string group_id, std::string stream) override;

    Error SetLastReadMarker(std::string group_id, uint64_t value, std::string stream) override;

    Error GetNext(std::string group_id, MessageMeta* info, MessageData* data, std::string stream) override;
    Error GetNext(std::string group_id, MessageMeta* info, MessageData* data, std::string stream, bool ordered) override;

    Error GetLast(MessageMeta* info, MessageData* data, std::string stream) override;
    Error GetLast(std::string group_id, MessageMeta* info, MessageData* data, std::string stream) override;

    std::string GenerateNewGroupId(Error* err) override;
    std::string GetBeamtimeMeta(Error* err) override;
    std::string GetStreamMeta(const std::string& stream, Error* err) override;

    uint64_t GetCurrentSize(std::string stream, Error* err) override;
    uint64_t GetCurrentDatasetCount(std::string stream, bool include_incomplete, Error* err) override;

    Error GetById(uint64_t id, MessageMeta* info, MessageData* data, std::string stream) override;

    Error GetVersionInfo(std::string* client_info, std::string* server_info, bool* supported) override;
    Error DeleteStream(std::string stream, DeleteStreamOptions options) override;
    Error SetStreamPersistent(std::string stream) override;
    void SetTimeout(uint64_t timeout_ms) override;
    void ForceNoRdma() override;
    Error DisableMonitoring(bool disable) override;

    NetworkConnectionType CurrentConnectionType() const override;

    MessageMetas QueryMessages(std::string query, std::string stream, Error* err) override;

    DataSet GetNextDataset(std::string group_id, uint64_t min_size, std::string stream, Error* err) override;
    DataSet GetNextDataset(std::string group_id, uint64_t min_size, std::string stream, bool ordered, Error* err) override;

    DataSet GetLastDataset(uint64_t min_size, std::string stream, Error* err) override;
    DataSet GetLastDataset(std::string group_id, uint64_t min_size, std::string stream, Error* err) override;

    DataSet GetDatasetById(uint64_t id, uint64_t min_size, std::string stream, Error* err) override;

    Error RetrieveData(MessageMeta* info, MessageData* data) override;

    StreamInfos GetStreamList(std::string from, StreamFilter filter, bool detailed, Error* err) override;
    StreamInfos GetStreamList(std::string from, StreamFilter filter, Error* err);
    SourceList GetSourceList(Error* err);

    void SetResendNacs(bool resend, uint64_t delay_ms, uint64_t resend_attempts) override;

    virtual void InterruptCurrentOperation() override;


    std::unique_ptr<IO> io__; // modified in testings to mock system calls,otherwise do not touch
    std::unique_ptr<NetClient> net_client__;
    std::unique_ptr<ServiceRequest> service_request_;
    std::mutex net_client_mutex__; // Required for the lazy initialization of net_client
  private:
    Error GetDataFromFileTransferService(MessageMeta* info, MessageData* data, bool retry_with_new_token);
    Error GetDataFromFile(MessageMeta* info, MessageData* data);
    static const std::string kBrokerServiceName;
    static const std::string kFileTransferServiceName;
    Error GetRecordFromServer(std::string* info, std::string group_id, std::string stream, GetMessageServerOperation op,
                              bool dataset = false, uint64_t min_size = 0);
    Error GetRecordFromServerById(uint64_t id, std::string* info, std::string group_id, std::string stream,
                                  bool dataset = false, uint64_t min_size = 0);
    Error GetDataIfNeeded(MessageMeta* info, MessageData* data);
    Error DiscoverService(const std::string& service_name, std::string* uri_to_set);
    bool SwitchToGetByIdIfNoData(Error* err, const std::string& response, std::string* group_id, std::string* redirect_uri);
    bool SwitchToGetByIdIfPartialData(Error* err, const std::string& response, std::string* group_id,
                                      std::string* redirect_uri);
    Error GetMessageFromServer(GetMessageServerOperation op, uint64_t id, std::string group_id, std::string stream,
                               MessageMeta* info, MessageData* data);
    DataSet GetDatasetFromServer(GetMessageServerOperation op, uint64_t id, std::string group_id, std::string stream,
                                 uint64_t min_size, Error* err);
    bool DataCanBeInBuffer(const MessageMeta* info);
    Error TryGetDataFromBuffer(const MessageMeta* info, MessageData* data);
    Error CreateNetClientAndTryToGetFile(const MessageMeta* info, const std::string& request_sender_details, MessageData* data);
    Error FtsRequestWithTimeout(MessageMeta* info, MessageData* data);
    Error FtsSizeRequestWithTimeout(MessageMeta* info);
    std::string OpToUriCmd(GetMessageServerOperation op);
    Error UpdateFolderTokenIfNeeded(bool ignore_existing);

    uint64_t GetCurrentCount(const RequestInfo& ri, Error* err);
    RequestInfo GetStreamListRequest(const std::string& from, const StreamFilter& filter, bool detailed) const;

    std::string endpoint_;
    std::string current_broker_uri_;
    std::string current_fts_uri_;
    std::string source_path_;
    bool has_filesystem_;
    SourceCredentials source_credentials_;
    bool use_new_api_format_ = true;
    std::string request_sender_details_prefix_;
    uint64_t timeout_ms_ = 0;
    bool should_try_rdma_first_ = true;
    NetworkConnectionType current_connection_type_ = NetworkConnectionType::kUndefined;
    std::string folder_token_;
    RequestInfo CreateFolderTokenRequest() const;
    RequestInfo CreateFileTransferRequest(const MessageMeta* info) const;
    bool resend_ = false;
    uint64_t delay_ms_;
    uint64_t resend_attempts_;
    std::atomic<bool> interrupt_flag_{ false};

    RequestInfo GetSizeRequestForSingleMessagesStream(std::string& stream) const;
    RequestInfo GetSizeRequestForDatasetStream(std::string& stream, bool include_incomplete) const;
    uint64_t ParseGetCurrentCountResponce(Error* err, const std::string& responce) const;
    RequestInfo GetDeleteStreamRequest(std::string stream, DeleteStreamOptions options) const;
};

}
#endif //ASAPO_CONSUMER_IMPL_H
