import {groupDelayStore} from "./groupDelayStore";

interface HitMissDataPoint {
    timestamp: number;
    hit: number;
    miss: number;
}

interface TransferTimePoint {
    timestamp: number;
    fromProducerToReceiverMs: number;
    receiverStaleTimeMs: number;
    fromReceiverToConsumerMs: number;
}

interface TotalTransferSpeedPoint {
    timestamp: number;
    avgBytesPerSecRecv: number;
    avgBytesPerSecSend: number;
}

interface TotalCacheMemoryUsagePoint {
    timestamp: number;
    totalUntouchedMemory: number;
    totalTouchedMemory: number;
}

export interface ToplogyStage {
    sourceName: string;
    totalFileRateRequested: number;
    totalGbRateRequested: number;

    receivers: ReceiverInfo[];
}

export interface ReceiverInfo {
    receiverId: string;
    fileRateIncoming: number;
    gbRateIncoming: number;
    usedMemory: number;
    totalMemory: number;
}

import {
    DataPointsQuery,
    DataPointsResponse,
    GroupDelayQuery,
    MetadataResponse,
    PipelineConnectionQueryMode,
    ToplogyQuery
} from '../generated_proto/AsapoMonitoringQueryService_pb';
import { Empty } from '../generated_proto/AsapoMonitoringCommonService_pb';
import { timeStore } from './timeStore';
import { selectionFilterStore } from './selectionFilterStore';

interface ConnectionStoreState {
    isFetchingMetaData: boolean;
    isFetchingData: boolean;
    isFetchingToplogyData: boolean;
    isFetchingGroupDelayData: boolean;

    // metadata
    clusterName: string;
    availableBeamtimes: string[];
}


import { reactive } from 'vue';
import { AsapoMonitoringQueryServiceClient } from '../generated_proto/AsapoMonitoringQueryService_grpc_web_pb';
import { errorStore } from './errorStore';
import { toplogyStore } from './toplogyStore';
import { metricData } from './metricDataStore';

console.log('AsapoMonitoringQueryServicePromiseClient');
const client = new AsapoMonitoringQueryServiceClient('/tv-api', null);

class ConnectionStore {
    private pendingDataRefreshTimeoutHandle?: number;

    private internalState: ConnectionStoreState = reactive({
        isFetchingMetaData: false,
        isFetchingData: false,
        isFetchingToplogyData: false,
        isFetchingGroupDelayData: false,

        clusterName: '<NOT SET>',
        availableBeamtimes: [],
    });

    public get state(): Readonly<ConnectionStoreState> {
        return this.internalState;
    }

    public setMetadata(initialData: MetadataResponse): void {
        this.internalState.clusterName = initialData.getClustername();
        this.internalState.availableBeamtimes = initialData.getAvailablebeamtimesList();
    }

    public setIsFetchingData(mode: boolean): void {
        this.internalState.isFetchingData = mode;
    }

    public setIsFetchingMetaData(mode: boolean): void {
        this.internalState.isFetchingMetaData = mode;
    }

    public setIsFetchingToplogyData(mode: boolean): void {
        this.internalState.isFetchingToplogyData = mode;
    }

    public setIsFetchingGroupDelayData(mode: boolean): void {
        this.internalState.isFetchingGroupDelayData = mode;
    }

    public triggerToplogyRefresh(): void {
        if (this.state.isFetchingToplogyData) {
            console.warn('Unable to fetch toplogy: Already fetching toplogy data.');
            return; // already fetching data
        }

        if (!selectionFilterStore.isValidFilterOrSetError()) {
            return;
        }

        this.setIsFetchingToplogyData(true);

        const query = new ToplogyQuery();

        const range = timeStore.state.lastFixedTimeRange;
        query.setFromtimestamp(range.fromUnixSec);
        query.setTotimestamp(range.toUnixSec);
        query.setBeamtimefilter(selectionFilterStore.state.beamtime!);

        errorStore.clearToplogyConnectionError();
        client.getTopology(query, undefined, (err, toplogyData) => {
            if (err) {
                errorStore.setToplogyConnectionErrorText(err.message);
                console.error('getTopology error: ', err);
                this.setIsFetchingToplogyData(false);
                return;
            }

            try {
                toplogyStore.parseServerResponse(toplogyData);
            }
            catch(e) {
                errorStore.setToplogyConnectionErrorText('internal: ' + e.message);
                console.error('getTopology internal error: ', e);
            }
            finally {
                this.setIsFetchingToplogyData(false);
            }

        });
    }

    public triggerGroupDelayRefresh(): void {
        if (this.state.isFetchingGroupDelayData) {
            console.warn('Unable to fetch group delay: Already fetching group delay data.');
            return; // already fetching data
        }

        if (!selectionFilterStore.isValidFilterOrSetError()) {
            return;
        }

        this.setIsFetchingGroupDelayData(true);

        const query = new GroupDelayQuery();

        const range = timeStore.state.lastFixedTimeRange;
        query.setFromtimestamp(range.fromUnixSec);
        query.setTotimestamp(range.toUnixSec);
        query.setBeamtimefilter(selectionFilterStore.state.beamtime!);

        errorStore.clearGroupDelayConnectionError();
        client.getGroupDelay(query, undefined, (err, groupDelayData) => {
            if (err) {
                errorStore.setGroupDelayConnectionErrorText(err.message);
                console.error('getGroupDelay error: ', err);
                this.setIsFetchingGroupDelayData(false);
                return;
            }

            try {
                groupDelayStore.parseServerResponse(groupDelayData);
            }
            catch(e) {
                errorStore.setGroupDelayConnectionErrorText('internal: ' + e.message);
                console.error('getGroupDelay internal error: ', e);
            }
            finally {
                this.setIsFetchingGroupDelayData(false);
            }

        });
    }

    public triggerMetaDataRefresh(): void {
        if (this.state.isFetchingMetaData) {
            console.warn('Unable to fetch metadata: Already fetching metadata.');
            return; // already fetching data
        }

        this.setIsFetchingMetaData(true);

        errorStore.clearMetadataConnectionError();
        client.getMetadata(new Empty(), undefined, (err, metaData) => {
            if (err) {
                errorStore.setMetadataConnectionErrorText(err.message);
                console.error('getMetadata error: ', err);
                this.setIsFetchingMetaData(false);
                return;
            }

            this.setMetadata(metaData);
            this.setIsFetchingMetaData(false);
        });
    }

    public clearPendingMetricsTimeout(): void {
        if (this.pendingDataRefreshTimeoutHandle !== undefined) { // remove pending delay
            clearTimeout(this.pendingDataRefreshTimeoutHandle);
            this.pendingDataRefreshTimeoutHandle = undefined;
        }
    }

    public triggerMetricsRefresh(reevaluateTime = true): void {
        if (this.state.isFetchingData) {
            console.warn('Unable to fetch data: Already fetching data.');
            return; // already fetching data
        }
        this.clearPendingMetricsTimeout();
        if (!selectionFilterStore.isValidFilterOrSetError()) {
            return;
        }

        if (reevaluateTime) {
            if (!timeStore.evaluateAndVerify()) {
                return;
            }
        }

        if (toplogyStore.couldBeOutdated()) {
            this.triggerToplogyRefresh();
        }

        if (groupDelayStore.couldBeOutdated()) {
            this.triggerGroupDelayRefresh();
        }

        this.setIsFetchingData(true);

        const query = new DataPointsQuery();

        query.setFromtimestamp(timeStore.state.lastFixedTimeRange.fromUnixSec);
        query.setTotimestamp(timeStore.state.lastFixedTimeRange.toUnixSec);

        if (selectionFilterStore.state.beamtime) {
            query.setBeamtimefilter(selectionFilterStore.state.beamtime);
        }
        if (selectionFilterStore.state.source) {
            query.setSourcefilter(selectionFilterStore.state.source);
        }
        if (selectionFilterStore.state.fromPipelineStepId) {
            query.setFrompipelinestepfilter(selectionFilterStore.state.fromPipelineStepId);
        }
        if (selectionFilterStore.state.toPipelineStepId) {
            query.setTopipelinestepfilter(selectionFilterStore.state.toPipelineStepId);
        }
        if (selectionFilterStore.state.fromPipelineStepId || selectionFilterStore.state.toPipelineStepId) {
            query.setPipelinequerymode(selectionFilterStore.state.pipelineFilterRelation === 'and' ? PipelineConnectionQueryMode.EXACTPATH : PipelineConnectionQueryMode.JUSTRELATEDTOSTEP);
        }
        if (selectionFilterStore.state.receiverId) {
            query.setReceiverfilter(selectionFilterStore.state.receiverId);
        }

        errorStore.clearMetricsConnectionError();
        client.getDataPoints(query, undefined, (err, data) => {
            if (err) {
                errorStore.setMetricsConnectionErrorText(err.message);
                console.error('getDataPoints error: ', err);
                this.setIsFetchingData(false);
                return;
            }

            metricData.parseServerResponse(data);
            if (timeStore.state.refreshIntervalInSec) {
                this.pendingDataRefreshTimeoutHandle = setTimeout(() => this.triggerMetricsRefresh(), timeStore.state.refreshIntervalInSec * 1000);
            }
            this.setIsFetchingData(false);
        });
    }
}

// Singelton instance
export const connection = new ConnectionStore();

connection.triggerMetaDataRefresh();
if (selectionFilterStore.state.beamtime) {
    connection.triggerMetricsRefresh();
    connection.triggerToplogyRefresh();
    connection.triggerGroupDelayRefresh();
}
