#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include "asapo/unittests/MockLogger.h"
#include "asapo/common/error.h"
#include "asapo/producer/common.h"
#include "../src/producer_impl.h"
#include "asapo/producer/producer_error.h"
#include "asapo/common/internal/version.h"

#include "../src/request_handler_tcp.h"
#include "asapo/request/request_pool_error.h"
#include "asapo/http_request/consumer_error.h"
#include "asapo/unittests/MockHttpClient.h"
#include "mocking.h"

namespace {

using ::testing::Return;
using ::testing::_;
using ::testing::DoAll;
using ::testing::SetArgReferee;
using ::testing::Gt;
using ::testing::Eq;
using ::testing::Ne;
using ::testing::Mock;
using ::testing::NiceMock;
using ::testing::InSequence;
using ::testing::HasSubstr;
using testing::SetArgPointee;


using asapo::RequestPool;
using asapo::ProducerRequest;
using asapo::MockHttpClient;
using asapo::HttpCode;


MATCHER_P10(M_CheckSendRequest, op_code, source_credentials, metadata, file_id, file_size, message, stream,
            ingest_mode,
            dataset_id,
            dataset_size,
            "Checks if a valid GenericRequestHeader was Send") {
    auto request = static_cast<ProducerRequest*>(arg);
    return ((asapo::GenericRequestHeader) (request->header)).op_code == op_code
           && ((asapo::GenericRequestHeader) (request->header)).data_id == file_id
           && ((asapo::GenericRequestHeader) (request->header)).data_size == uint64_t(file_size)
           && request->manage_data_memory == true
           && request->source_credentials == source_credentials
           && request->metadata == metadata
           && (op_code == asapo::kOpcodeTransferMetaData ? ((asapo::GenericRequestHeader) (request->header)).custom_data[1]
               == uint64_t(dataset_id) : true)
           && (op_code == asapo::kOpcodeTransferDatasetData ? ((asapo::GenericRequestHeader) (request->header)).custom_data[1]
               == uint64_t(dataset_id) : true)
           && (op_code == asapo::kOpcodeTransferDatasetData ? ((asapo::GenericRequestHeader) (request->header)).custom_data[2]
               == uint64_t(dataset_size) : true)
           && ((asapo::GenericRequestHeader) (request->header)).custom_data[asapo::kPosIngestMode] == uint64_t(ingest_mode)
           && strcmp(((asapo::GenericRequestHeader) (request->header)).message, message) == 0
           && strcmp(((asapo::GenericRequestHeader) (request->header)).api_version,
                     asapo::kProducerProtocol.GetReceiverVersion().c_str()) == 0
           && strcmp(((asapo::GenericRequestHeader) (request->header)).stream, stream) == 0;


}

TEST(ProducerImpl, Constructor) {
    asapo::ProducerImpl producer{"", 4, 3600000, asapo::RequestHandlerType::kTcp};
    ASSERT_THAT(dynamic_cast<asapo::AbstractLogger*>(producer.log__), Ne(nullptr));
    ASSERT_THAT(dynamic_cast<asapo::RequestPool*>(producer.request_pool__.get()), Ne(nullptr));
}

class ProducerImplTests : public testing::Test {
  public:
    testing::NiceMock<MockDiscoveryService> service;
    asapo::ProducerRequestHandlerFactory factory{&service};
    testing::NiceMock<asapo::MockLogger> mock_logger;
    testing::NiceMock<MockRequestPool> mock_pull{&factory, &mock_logger};
    std::string expected_server_uri = "test:8400";
    asapo::ProducerImpl producer{expected_server_uri, 1, 3600000, asapo::RequestHandlerType::kTcp};
    uint64_t expected_size = 100;
    uint64_t expected_id = 10;
    uint64_t expected_dataset_id = 100;
    uint64_t expected_dataset_size = 4;
    uint64_t expected_ingest_mode = asapo::IngestModeFlags::kTransferMetaDataOnly;
    uint64_t request_timeout = 100;

    std::string expected_token = "token";
    std::string expected_broker_uri = "asapo-broker:5005";
    std::string expected_consumer_protocol = asapo::kConsumerProtocol.GetVersion();
    std::string expected_broker_protocol = asapo::kConsumerProtocol.GetBrokerVersion();
    std::string expected_broker_api = expected_broker_uri + "/" + expected_broker_protocol;
    std::string expected_data_source_encoded = "source%2F%24.%3F";
    std::string expected_stream_encoded = "str%20%24%20eam%24";
    std::string expected_stream_info = "{\"lastId\":3,\"name\":\"stream name\",\"timestampCreated\":1716811188757426483,\"timestampLast\":1716811188792663748,\"finished\":false,\"nextStream\":\"\"}";


    std::string expected_group_id = "groupid$";
    std::string expected_group_id_encoded = "groupid%24";
    std::string expected_data_source = "source/$.?";
    std::string expected_instance_id = "some instance";
    std::string expected_pipeline_step = "a new step";
    std::string expected_instance_id_encoded = "some%20instance";
    std::string expected_pipeline_step_encoded = "a%20new%20step";


    char expected_name[asapo::kMaxMessageSize] = "test_name";
    char expected_stream[asapo::kMaxMessageSize] = "test_stream";
    std::string expected_next_stream = "next_stream";

    asapo::SourceCredentials expected_credentials{asapo::SourceType::kRaw, "instance", "step", "beamtime_id", "beamline", "subname", "token"};
    asapo::SourceCredentials expected_default_credentials{asapo::SourceType::kProcessed, "instance", "step", "beamtime_id", "", "", "token"};

    std::string expected_credentials_str = "raw%instance%step%beamtime_id%beamline%subname%token";
    std::string expected_default_credentials_str = "processed%instance%step%beamtime_id%auto%detector%token";

    std::string expected_metadata = "meta";
    std::string expected_fullpath = "filename";
    bool expected_managed_memory = true;
    bool expected_unmanaged_memory = false;

    NiceMock<MockHttpClient> mock_http_client{};

    void SetUp() override {
        producer.log__ = &mock_logger;
        producer.request_pool__ = std::unique_ptr<RequestPool> {&mock_pull};
        producer.SetCredentials(expected_credentials);
        producer.service_request_->httpclient__ = std::unique_ptr<asapo::HttpClient> {&mock_http_client};

        {
            ON_CALL(mock_http_client, UrlEscape_t(expected_instance_id)).WillByDefault(Return(expected_instance_id_encoded));
            ON_CALL(mock_http_client, UrlEscape_t(expected_pipeline_step)).WillByDefault(Return(expected_pipeline_step_encoded));
            ON_CALL(mock_http_client, UrlEscape_t(expected_stream)).WillByDefault(Return(expected_stream_encoded));
            ON_CALL(mock_http_client, UrlEscape_t(expected_group_id)).WillByDefault(Return(expected_group_id_encoded));
            ON_CALL(mock_http_client, UrlEscape_t(expected_data_source)).WillByDefault(Return(expected_data_source_encoded));
            ON_CALL(mock_http_client, UrlEscape_t("0")).WillByDefault(Return("0"));
            ON_CALL(mock_http_client, UrlEscape_t("")).WillByDefault(Return(""));
            ON_CALL(mock_http_client, UrlEscape_t("default")).WillByDefault(Return("default"));
            ON_CALL(mock_http_client, UrlEscape_t("stream")).WillByDefault(Return("stream"));
            ON_CALL(mock_http_client, UrlEscape_t("instance")).WillByDefault(Return("instance"));
            ON_CALL(mock_http_client, UrlEscape_t("step")).WillByDefault(Return("step"));
            ON_CALL(mock_http_client, UrlEscape_t("DefaultStep")).WillByDefault(Return("DefaultStep"));
            ON_CALL(mock_http_client, UrlEscape_t("a")).WillByDefault(Return("b"));
            ON_CALL(mock_http_client, UrlEscape_t("b")).WillByDefault(Return("b"));
        }

    }
    void TearDown() override {
        producer.request_pool__.release();
        producer.service_request_.release();

        Mock::VerifyAndClear(&mock_http_client);
    }

    void MockGetServiceUri(std::string service, std::string result) {
        EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expected_server_uri + "/asapo-discovery/v0.1/" + service + "?token="
                                                      + expected_token + "&protocol=" + expected_consumer_protocol), _, _)
                                            ).WillOnce(DoAll(
                                                    SetArgPointee<1>(HttpCode::OK),
                                                    SetArgPointee<2>(nullptr),
                                                    Return(result)));
    }

    void MockGetBrokerUri() {
        MockGetServiceUri("asapo-broker", expected_broker_uri);
    }

};

TEST_F(ProducerImplTests, SendReturnsError) {
    EXPECT_CALL(mock_pull, AddRequest_t(_, false)).WillOnce(Return(
                asapo::IOErrorTemplates::kNoSpaceLeft.Generate().release()));
    asapo::MessageHeader message_header{1, 1, "test"};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kRequestPoolIsFull));
}

TEST_F(ProducerImplTests, ErrorIfFileNameTooLong) {
    asapo::MessageData data = asapo::MessageData{new uint8_t[100]};
    data[34] = 12;
    std::string long_string(asapo::kMaxMessageSize + 100, 'a');
    asapo::MessageHeader message_header{1, 1, long_string};
    auto err = producer.Send(message_header, std::move(data), expected_ingest_mode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
    auto err_data = static_cast<asapo::OriginalData*>(err->GetCustomData());
    ASSERT_THAT(err_data, Ne(nullptr));
    ASSERT_THAT(err_data->data[34], Eq(12));
}

TEST_F(ProducerImplTests, ErrorIfStreamEmpty) {
    asapo::MessageHeader message_header{1, 100, expected_fullpath};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, "", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}


TEST_F(ProducerImplTests, ErrorIfFileEmpty) {
    std::string long_string(asapo::kMaxMessageSize + 100, 'a');
    asapo::MessageHeader message_header{1, 1, ""};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorIfDatasetSizeNotDefined) {
    EXPECT_CALL(mock_logger, Error(testing::HasSubstr("dataset dimensions")));
    asapo::MessageHeader message_header{1, 1000, "test", "", 1};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorIfZeroDataSize) {
    asapo::MessageData data = asapo::MessageData{new uint8_t[100]};
    asapo::MessageHeader message_header{1, 0, expected_fullpath};
    auto err = producer.Send(message_header, std::move(data), asapo::kDefaultIngestMode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
    auto err_data = static_cast<asapo::OriginalData*>(err->GetCustomData());
    ASSERT_THAT(err_data, Ne(nullptr));
}

TEST_F(ProducerImplTests, ErrorIfBothIdAndAutoIdSet) {
    asapo::MessageData data = asapo::MessageData{new uint8_t[100]};
    asapo::MessageHeader message_header{1, 100, expected_fullpath, "", 0, 0, true};
    auto err = producer.Send(message_header, std::move(data), asapo::kDefaultIngestMode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
    auto err_data = static_cast<asapo::OriginalData*>(err->GetCustomData());
    ASSERT_THAT(err_data, Ne(nullptr));
}

TEST_F(ProducerImplTests, OkAutoId) {
    asapo::MessageHeader message_header{0, 100, expected_fullpath, "", 0, 0, true};
    auto err = producer.Send(message_header, nullptr, asapo::kTransferMetaDataOnly, "default", nullptr);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, ErrorIfAutoIdForSet) {
    asapo::MessageHeader message_header{0, 0, expected_fullpath, "", 1, 1, true};
    auto err = producer.Send(message_header, nullptr, asapo::kTransferMetaDataOnly, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorIfNoData) {
    asapo::MessageHeader message_header{1, 100, expected_fullpath};
    auto err = producer.Send(message_header, nullptr, asapo::kDefaultIngestMode, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorIfNoDataSend_) {
    asapo::MessageHeader message_header{1, 100, expected_fullpath};
    auto err = producer.Send__(message_header, nullptr, asapo::kDefaultIngestMode, expected_stream, nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorIfSendingDataWithZeroId) {
    asapo::MessageHeader message_header{0, 100, expected_fullpath};
    auto err = producer.Send(message_header, nullptr, asapo::kTransferMetaDataOnly, "default", nullptr);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, OkIfNoDataWithTransferMetadataOnlyMode) {
    asapo::MessageHeader message_header{1, 100, expected_fullpath};
    auto err = producer.Send(message_header, nullptr, asapo::kTransferMetaDataOnly, "default", nullptr);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OkIfZeroSizeWithTransferMetadataOnlyMode) {
    asapo::MessageData data = asapo::MessageData{new uint8_t[100]};
    asapo::MessageHeader message_header{1, 0, expected_fullpath};
    auto err = producer.Send(message_header, std::move(data), asapo::kTransferMetaDataOnly, "default", nullptr);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKSendingSendRequestWithStream) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferData,
                                                           expected_credentials_str,
                                                           expected_metadata,
                                                           expected_id,
                                                           expected_size,
                                                           expected_name,
                                                           expected_stream,
                                                           expected_ingest_mode,
                                                           0,
                                                           0
                                                           ), false)).WillOnce(Return(
                                                                   nullptr));

    asapo::MessageHeader message_header{expected_id, expected_size, expected_name, expected_metadata};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKSendingSendRequestWithStream_PreSet) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferData,
                                                           expected_credentials_str,
                                                           expected_metadata,
                                                           expected_id,
                                                           expected_size,
                                                           expected_name,
                                                           expected_stream,
                                                           expected_ingest_mode,
                                                           0,
                                                           0
                                                           ), false)).WillOnce(Return(
                                                                   nullptr));

    asapo::MessageHeader message_header{expected_id, expected_size, expected_name, expected_metadata};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKSendingStreamFinish) {
    producer.SetCredentials(expected_credentials);

    std::string next_stream_meta = std::string("{\"next_stream\":") + "\"" + expected_next_stream + "\"}";

    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferData,
                                        expected_credentials_str,
                                        next_stream_meta.c_str(),
                                        expected_id + 1,
                                        0,
                                        asapo::kFinishStreamKeyword.c_str(),
                                        expected_stream,
                                        asapo::IngestModeFlags::kTransferMetaDataOnly,
                                        0,
                                        0
                                                          ), false)).WillOnce(Return(
                                                                  nullptr));

    auto err = producer.SendStreamFinishedFlag(expected_stream, expected_id, expected_next_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, ErrorSendingStreamFinishWithemptyStream) {
    producer.SetCredentials(expected_credentials);
    auto err = producer.SendStreamFinishedFlag("", expected_id, expected_next_stream, nullptr);

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, OKSendingStreamFinishWithNoNextStream) {
    producer.SetCredentials(expected_credentials);

    std::string
    next_stream_meta = std::string("{\"next_stream\":") + "\"" + asapo::kNoNextStreamKeyword
                       + "\"}";

    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferData,
                                        expected_credentials_str,
                                        next_stream_meta.c_str(),
                                        expected_id + 1,
                                        0,
                                        asapo::kFinishStreamKeyword.c_str(),
                                        expected_stream,
                                        asapo::IngestModeFlags::kTransferMetaDataOnly,
                                        0,
                                        0
                                                          ), false)).WillOnce(Return(
                                                                  nullptr));

    auto err = producer.SendStreamFinishedFlag(expected_stream, expected_id, "", nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKSendingSendDatasetDataRequest) {
    producer.SetCredentials(expected_credentials);
    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferDatasetData,
                                        expected_credentials_str,
                                        expected_metadata,
                                        expected_id,
                                        expected_size,
                                        expected_name,
                                        expected_stream,
                                        expected_ingest_mode,
                                        expected_dataset_id,
                                        expected_dataset_size), false)).WillOnce(
                                            Return(
                                                nullptr));

    asapo::MessageHeader message_header
    {expected_id, expected_size, expected_name, expected_metadata, expected_dataset_id, expected_dataset_size};
    auto err = producer.Send(message_header, nullptr, expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKAddingSendMetaDataRequestOld) {
    expected_id = 0;
    expected_metadata = "{\"meta\":10}";
    expected_size = expected_metadata.size();
    expected_ingest_mode = asapo::IngestModeFlags::kTransferData | asapo::IngestModeFlags::kStoreInDatabase ;

    producer.SetCredentials(expected_credentials);
    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferMetaData,
                                        expected_credentials_str,
                                        "",
                                        expected_id,
                                        expected_size,
                                        "beamtime_global.meta",
                                        "",
                                        expected_ingest_mode,
                                        12,
                                        10), false)).WillOnce(Return(
                                                    nullptr));

    auto err = producer.SendMetadata(expected_metadata, nullptr);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKAddingSendMetaDataRequest) {
    expected_id = 0;
    expected_metadata = "{\"meta\":10}";
    expected_size = expected_metadata.size();
    expected_ingest_mode = asapo::IngestModeFlags::kTransferData | asapo::IngestModeFlags::kStoreInDatabase ;

    producer.SetCredentials(expected_credentials);
    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferMetaData,
                                        expected_credentials_str,
                                        "",
                                        expected_id,
                                        expected_size,
                                        "beamtime_global.meta",
                                        "",
                                        expected_ingest_mode,
                                        12,
                                        10), false)).WillOnce(Return(
                                                    nullptr));

    auto mode = asapo::MetaIngestMode{asapo::MetaIngestOp::kReplace, true};
    auto err = producer.SendBeamtimeMetadata(expected_metadata, mode, nullptr);
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, OKAddingSendStreamDataRequest) {
    expected_id = 0;
    expected_metadata = "{\"meta\":10}";
    expected_size = expected_metadata.size();
    expected_ingest_mode = asapo::IngestModeFlags::kTransferData | asapo::IngestModeFlags::kStoreInDatabase ;

    std::string expected_message = (std::string(expected_stream) + ".meta");
    producer.SetCredentials(expected_credentials);
    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferMetaData,
                                        expected_credentials_str,
                                        "",
                                        expected_id,
                                        expected_size,
                                        expected_message.c_str(),
                                        expected_stream,
                                        expected_ingest_mode,
                                        1,
                                        10), false)).WillOnce(Return(
                                                    nullptr));


    auto mode = asapo::MetaIngestMode{asapo::MetaIngestOp::kInsert, false};
    auto err = producer.SendStreamMetadata(expected_metadata, mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}


TEST_F(ProducerImplTests, ErrorSendingEmptyFileName) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(_, _)).Times(0);

    asapo::MessageHeader message_header{expected_id, 0, expected_name};
    auto err = producer.SendFile(message_header, "", expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));

}

TEST_F(ProducerImplTests, ErrorSendingEmptyRelativeFileName) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(_, _)).Times(0);

    asapo::MessageHeader message_header{expected_id, 0, ""};
    auto err = producer.SendFile(message_header, expected_fullpath, expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));

}

TEST_F(ProducerImplTests, ErrorSendingFileToEmptyStream) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(_, _)).Times(0);

    asapo::MessageHeader message_header{expected_id, 0, expected_name};
    auto err = producer.SendFile(message_header, expected_fullpath, expected_ingest_mode, "", nullptr);

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));

}

TEST_F(ProducerImplTests, OKSendingSendFileRequestWithStream) {
    producer.SetCredentials(expected_credentials);

    EXPECT_CALL(mock_pull, AddRequest_t(M_CheckSendRequest(asapo::kOpcodeTransferData,
                                        expected_credentials_str,
                                        "",
                                        expected_id,
                                        0,
                                        expected_name,
                                        expected_stream,
                                        expected_ingest_mode,
                                        0,
                                        0), false)).WillOnce(Return(
                                                    nullptr));

    asapo::MessageHeader message_header{expected_id, 0, expected_name};
    auto err =
        producer.SendFile(message_header, expected_fullpath, expected_ingest_mode, expected_stream, nullptr);

    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(ProducerImplTests, ErrorSettingSecondTime) {
    EXPECT_CALL(mock_logger, Error(testing::HasSubstr("already")));

    auto err = producer.SetCredentials(asapo::SourceCredentials{asapo::SourceType::kRaw, "instance", "step", "4", "", "5", "6"});

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}

TEST_F(ProducerImplTests, ErrorSendingWrongIngestMode) {
    producer.SetCredentials(expected_credentials);
    asapo::MessageHeader message_header{expected_id, 0, expected_name};
    uint64_t ingest_modes[] = {0, asapo::IngestModeFlags::kTransferMetaDataOnly | asapo::IngestModeFlags::kTransferData,
                               asapo::IngestModeFlags::kTransferData,
                               asapo::IngestModeFlags::kStoreInDatabase,
                               asapo::IngestModeFlags::kStoreInFilesystem,
                               asapo::IngestModeFlags::kStoreInDatabase | asapo::IngestModeFlags::kStoreInFilesystem,
                               asapo::IngestModeFlags::kTransferMetaDataOnly
                               | asapo::IngestModeFlags::kStoreInFilesystem
                              };

    EXPECT_CALL(mock_pull, AddRequest_t(_, _)).Times(0);

    for (auto ingest_mode : ingest_modes) {
        auto err = producer.SendFile(message_header, expected_fullpath, ingest_mode, expected_stream, nullptr);
        ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
    }

}

TEST_F(ProducerImplTests, GetQueueSize) {
    EXPECT_CALL(mock_pull, NRequestsInPool()).WillOnce(Return(10));

    auto size = producer.GetRequestsQueueSize();
    ASSERT_THAT(size, Eq(10));
}

TEST_F(ProducerImplTests, GetQueueVolume) {
    EXPECT_CALL(mock_pull, UsedMemoryInPool()).WillOnce(Return(10000000));

    auto vol = producer.GetRequestsQueueVolumeMb();

    ASSERT_THAT(vol, Eq(10));
}

MATCHER_P(M_CheckLimits, limits, "Checks if a valid limits were used") {
    return arg.max_requests == limits.max_requests && arg.max_memory_mb == limits.max_memory_mb;
}

TEST_F(ProducerImplTests, SetLimits) {
    EXPECT_CALL(mock_pull, SetLimits(M_CheckLimits(asapo::RequestPoolLimits{10, 20})));

    producer.SetRequestsQueueLimits(10, 20);
}

TEST_F(ProducerImplTests, WaitRequestsFinished) {
    EXPECT_CALL(mock_pull, WaitRequestsFinished_t(_)).WillOnce(Return(
                asapo::IOErrorTemplates::kTimeout.Generate().release()));

    auto err = producer.WaitRequestsFinished(100);

    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kTimeout));
}

MATCHER_P3(M_CheckGetRequest, op_code, source_credentials, stream,
           "Checks if a valid GenericRequestHeader was Send") {
    auto request = static_cast<ProducerRequest*>(arg);
    return ((asapo::GenericRequestHeader) (arg->header)).op_code == op_code
           && request->source_credentials == source_credentials
           && strcmp(((asapo::GenericRequestHeader) (arg->header)).stream, stream) == 0;
}

TEST_F(ProducerImplTests, GetStreamInfoMakesCorrectRequest) {

    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client,
                Get_t(expected_broker_api + "/beamtime/beamtime_id/subname/0/streams?token=token&detailed=true&from=str%20%24%20eam%24&single=true", _,
                      _)).WillOnce(DoAll(
                                       SetArgPointee<1>(HttpCode::OK),
                                       SetArgPointee<2>(nullptr),
                                       Return("{\"streams\":[" + expected_stream_info + "]}")));

    asapo::Error err;
    auto res = producer.GetStreamInfo(expected_stream, request_timeout, &err);
    ASSERT_THAT(err, Eq(nullptr));
    ASSERT_THAT(res.Json(), Eq(expected_stream_info));

}


TEST_F(ProducerImplTests, GetStreamInfoTimeout) {
    asapo::Error err;
    auto sinfo = producer.GetStreamInfo("stream", request_timeout, &err);
    // Timeout error is happen because response is not mocked and therefore service is unavailable.
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kTimeout));
}


TEST_F(ProducerImplTests, GetStreamInfoErrorOnEmptyStream) {
    producer.SetCredentials(expected_credentials);
    asapo::Error err;
    producer.GetStreamInfo("", request_timeout, &err);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kWrongInput));
}


TEST_F(ProducerImplTests, GetLastStreamMakesCorrectRequest) {

    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client,
                Get_t(expected_broker_api + "/beamtime/beamtime_id/subname/0/streams?token=token&detailed=true&from=&last=true", _,
                      _)).WillOnce(DoAll(
                                       SetArgPointee<1>(HttpCode::OK),
                                       SetArgPointee<2>(nullptr),
                                       Return("{\"streams\":[" + expected_stream_info + "]}")));

    asapo::Error err;
    auto res = producer.GetLastStream(request_timeout, &err);
    ASSERT_THAT(err, Eq(nullptr));
    ASSERT_THAT(res.Json(), Eq(expected_stream_info));

}


TEST_F(ProducerImplTests, ReturnDataIfCanotAddToQueue) {
    producer.SetCredentials(expected_credentials);

    asapo::MessageData data = asapo::MessageData{new uint8_t[100]};
    data[40] = 10;
    asapo::OriginalRequest* original_request = new asapo::OriginalRequest{};

    auto request = std::unique_ptr<ProducerRequest> {new ProducerRequest{"", asapo::GenericRequestHeader{}, std::move(data), "", "", nullptr, true, 0}};
    original_request->request = std::move(request);
    auto pool_err = asapo::IOErrorTemplates::kNoSpaceLeft.Generate();
    pool_err->SetCustomData(std::unique_ptr<asapo::CustomErrorData> {original_request});


    EXPECT_CALL(mock_pull, AddRequest_t(_, _)).WillOnce(Return(
                std::move(pool_err).release()));

    asapo::MessageHeader message_header{expected_id, 0, expected_name};
    data = asapo::MessageData{new uint8_t[100]};
    auto err = producer.Send(message_header, std::move(data), expected_ingest_mode, expected_stream, nullptr);

    auto err_data = static_cast<asapo::OriginalData*>(err->GetCustomData());
    ASSERT_THAT(err_data, Ne(nullptr));

    asapo::MessageData original_data_in_err = std::move(err_data->data);
    ASSERT_THAT(err, Eq(asapo::ProducerErrorTemplates::kRequestPoolIsFull));
    ASSERT_THAT(original_data_in_err, Ne(nullptr));
    ASSERT_THAT(original_data_in_err[40], Eq(10));
}

TEST_F(ProducerImplTests, GetVersionInfoWithServer) {

    std::string result_producer =
        R"({"softwareVersion":"21.06.0, build 7a9294ad","clientSupported":"no", "clientProtocol":{"versionInfo":"v0.4"}})";

    std::string result_consumer =
        R"({"softwareVersion":"21.06.0, build 7a9294ad","clientSupported":"no", "clientProtocol":{"versionInfo":"v0.5"}})";

    EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expected_server_uri +
                                                   "/asapo-discovery/v0.1/version?token=token&client=producer&protocol=v0.6"), _, _)).WillOnce(DoAll(
                                                           SetArgPointee<1>(asapo::HttpCode::OK),
                                                           SetArgPointee<2>(nullptr),
                                                           Return(result_producer)));


    EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expected_server_uri +
                                                   "/asapo-discovery/v0.1/version?token=token&client=consumer&protocol=v0.6"), _, _)).WillOnce(DoAll(
                                                           SetArgPointee<1>(asapo::HttpCode::OK),
                                                           SetArgPointee<2>(nullptr),
                                                           Return(result_consumer)));

    std::string client_info, server_info;
    auto err = producer.GetVersionInfo(&client_info, &server_info, nullptr);
    ASSERT_THAT(err, Eq(nullptr));
    ASSERT_THAT(server_info, HasSubstr("21.06.0"));
    ASSERT_THAT(server_info, HasSubstr("v0.4"));
    ASSERT_THAT(server_info, HasSubstr("v0.5"));
}

MATCHER_P4(M_CheckDeleteStreamRequest, op_code, source_credentials, stream, flag,
           "Checks if a valid GenericRequestHeader was Send") {
    auto request = static_cast<ProducerRequest*>(arg);
    return ((asapo::GenericRequestHeader) (arg->header)).op_code == op_code
           && request->source_credentials == source_credentials
           && ((asapo::GenericRequestHeader) (arg->header)).custom_data[0] == static_cast<uint64_t>(flag)
           && strcmp(((asapo::GenericRequestHeader) (arg->header)).stream, stream) == 0;
}

TEST_F(ProducerImplTests, DeleteStreamMakesCorrectRequest) {
    MockGetBrokerUri();

    std::string expected_delete_stream_query_string = "{\"ErrorOnNotExist\":true,\"DeleteMeta\":true}";
    EXPECT_CALL(mock_http_client, Post_t(expected_broker_api + "/beamtime/beamtime_id/subname" + "/"
                                         + expected_stream_encoded + "/delete"
                                         + "?token=token&instanceid=instance&pipelinestep=step", _,
                                         expected_delete_stream_query_string, _, _)).WillOnce(DoAll(
                                                     SetArgPointee<3>(HttpCode::OK),
                                                     SetArgPointee<4>(nullptr),
                                                     Return("")
                                                 ));

    asapo::DeleteStreamOptions opt;
    opt.delete_meta = true;
    opt.error_on_not_exist = true;
    auto err = producer.DeleteStream(expected_stream, request_timeout, opt);
    ASSERT_THAT(err, Eq(nullptr));

}

TEST_F(ProducerImplTests, GetStreamMetaMakesCorrectRequest) {
    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client, Get_t(expected_broker_api + "/beamtime/beamtime_id/subname" +
                                        "/" + expected_stream_encoded + "/0/meta/1?token=token&instanceid=instance&pipelinestep=step", _,
                                        _)).WillOnce(DoAll(
                                                SetArgPointee<1>(HttpCode::OK),
                                                SetArgPointee<2>(nullptr),
                                                Return(expected_metadata)));

    asapo::Error err;
    auto res = producer.GetStreamMeta(expected_stream, request_timeout, &err);
    ASSERT_THAT(err, Eq(nullptr));
    ASSERT_THAT(res, Eq(expected_metadata));
}

TEST_F(ProducerImplTests, GetBeamtimeMetaMakesCorrectRequest) {
    std::string expected_metadata = "{\"meta\":1}";
    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client, Get_t(expected_broker_api + "/beamtime/beamtime_id/subname" +
                                        "/default/0/meta/0?token=token&instanceid=instance&pipelinestep=step", _,
                                        _)).WillOnce(DoAll(
                                                SetArgPointee<1>(HttpCode::OK),
                                                SetArgPointee<2>(nullptr),
                                                Return(expected_metadata)));

    asapo::Error err;
    auto res = producer.GetBeamtimeMeta(request_timeout, &err);
    ASSERT_THAT(err, Eq(nullptr));
    ASSERT_THAT(res, Eq(expected_metadata));
}

}
