#include <functional>

#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include "asapo/database/db_error.h"

#include "asapo/unittests/MockIO.h"
#include "asapo/unittests/MockDatabase.h"
#include "asapo/unittests/MockLogger.h"
#include "asapo/unittests/MockHttpClient.h"

#include "../../src/receiver_error.h"
#include "../../src/request.h"
#include "../../src/request_handler/request_factory.h"
#include "../../src/request_handler/request_handler.h"
#include "../../src/request_handler/request_handler_db_check_request.h"
#include "asapo/common/networking.h"
#include "../mock_receiver_config.h"
#include "asapo/common/data_structs.h"
#include "asapo/common/internal/version.h"

#include "../receiver_mocking.h"

using namespace testing;
using namespace asapo;

using MockFunctions = std::vector<std::function<void(asapo::ErrorInterface*, bool )>>;

namespace {


class DbCheckRequestHandlerTests : public Test {
  public:
    std::string expected_stream = "stream";
    std::string expected_collection_name = std::string(asapo::kDBDataCollectionNamePrefix) + "_" + expected_stream;
    RequestHandlerDbCheckRequest handler{};
    std::unique_ptr<NiceMock<MockRequest>> mock_request;
    NiceMock<asapo::MockLogger> mock_logger;
    ReceiverConfig config;
    std::string expected_beamtime_id = "beamtime_id";
    std::string expected_default_source = "detector";
    std::string expected_data_source = "source";
    std::string expected_host_uri = "127.0.0.1:1234";
    uint64_t expected_port = 1234;
    uint64_t expected_buf_id = 18446744073709551615ull;
    std::string expected_file_name = "2";
    std::string expected_metadata = "{\"meta\":1}";
    uint64_t expected_file_size = 10;
    uint64_t expected_id = 15;
    uint64_t expected_dataset_id = 16;
    uint64_t expected_dataset_size = 2;
    CustomRequestData expected_custom_data{0, expected_dataset_id, expected_dataset_size};
    MessageMeta expected_message_meta;
    int n_run = 0;

    std::string expected_token = "token";
    std::string expected_server_uri = "localhost:8400";
    std::string expected_broker_uri = "asapo-broker:5005";
    std::string expected_consumer_protocol = asapo::kConsumerProtocol.GetVersion();
    std::string expected_broker_protocol = asapo::kConsumerProtocol.GetBrokerVersion();
    std::string expected_broker_api = expected_broker_uri + "/" + expected_broker_protocol;
    std::string expected_stream_encoded = "str%20%24%20eam%24";

    std::string expected_pipeline_step_id = "pipeline_step_id";
    std::string expected_producer_instance_id = "producer_instance_id";

    NiceMock<MockHttpClient> mock_http_client{};
    void SetUp() override {
        GenericRequestHeader request_header;
        request_header.data_id = 2;
        request_header.op_code = asapo::Opcode::kOpcodeTransferData;
        handler.log__ = &mock_logger;
        handler.service_request_->httpclient__ = std::unique_ptr<asapo::HttpClient> {&mock_http_client};

        mock_request.reset(new NiceMock<MockRequest> {request_header, 1, "", nullptr, nullptr});

        ON_CALL(*mock_request, GetPipelineStepId()).WillByDefault(ReturnRef(expected_pipeline_step_id));
        ON_CALL(*mock_request, GetProducerInstanceId()).WillByDefault(ReturnRef(expected_producer_instance_id));
        ON_CALL(*mock_request, GetBeamtimeId()).WillByDefault(ReturnRef(expected_beamtime_id));
        ON_CALL(*mock_request, GetDataSource()).WillByDefault(ReturnRef(expected_data_source));
        ON_CALL(*mock_request, GetStream()).WillByDefault(Return(expected_stream));
        ON_CALL(*mock_request, GetFileName()).WillByDefault(Return(expected_file_name));
        ON_CALL(mock_http_client, UrlEscape_t(expected_stream)).WillByDefault(Return(expected_stream_encoded));

        config.database_uri = "127.0.0.1:27017";
        config.dataserver.advertise_uri = expected_host_uri;
        config.dataserver.listen_port = expected_port;
        SetReceiverConfig(config, "none");
        expected_message_meta =  PrepareMessageMeta();
        SetDefaultRequestCalls(mock_request.get(),expected_beamtime_id);

        ON_CALL(mock_http_client, UrlEscape_t(expected_stream)).WillByDefault(Return(expected_stream_encoded));
        ON_CALL(mock_http_client, UrlEscape_t("default")).WillByDefault(Return("default_encoded"));

    }
    void TearDown() override {
        handler.service_request_.release();
        Mock::VerifyAndClear(&mock_http_client);
        Mock::VerifyAndClearExpectations(mock_request.get());
    }
    void MockGetServiceUri(std::string service, std::string result) {
        EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expected_server_uri + "/asapo-discovery/v0.1/" + service + "?token="
                                                      + expected_token + "&protocol=" + expected_consumer_protocol), _, _)
                                            ).WillOnce(DoAll(
                                                    SetArgPointee<1>(HttpCode::OK),
                                                    SetArgPointee<2>(nullptr),
                                                    Return(result)));
    }

    void MockGetBrokerUri() {
        MockGetServiceUri("asapo-broker", expected_broker_uri);
    }

    void MockGetData(std::string response, asapo::HttpCode code = HttpCode::OK){
        auto expect_url =  expected_broker_api + "/beamtime/beamtime_id//"+expected_stream_encoded
                                            +"/0/" + std::to_string(expected_id)
                                            + "?token=token&instanceid=&pipelinestep=&id=" + std::to_string(expected_id);
        EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expect_url), _, _)
        ).WillOnce(DoAll(
                                                    SetArgPointee<1>(code),
                                                    SetArgPointee<2>(nullptr),
                                                    Return(response)));
    }

    void ExpectRequestParams(asapo::Opcode op_code, const std::string& data_source, bool expect_compare = true);

    MessageMeta PrepareMessageMeta();

};

MATCHER_P(CompareMessageMeta, file, "") {
    if (arg.size != file.size) return false;
    if (arg.source != file.source) return false;
    if (arg.buf_id != file.buf_id) return false;
    if (arg.name != file.name) return false;
    if (arg.id != file.id) return false;
    if (arg.metadata != file.metadata) return false;

    return true;
}


void DbCheckRequestHandlerTests::ExpectRequestParams(asapo::Opcode op_code, const std::string& data_source,
        bool expect_compare) {

    std::string db_name = expected_beamtime_id;
    db_name += "_" + data_source;

    if (n_run  == 0) {
        EXPECT_CALL(*mock_request, GetDataSource())
        .WillRepeatedly(ReturnRef(data_source))
        ;
    }

    if (expect_compare) {
        EXPECT_CALL(*mock_request, GetDataSize())
        .WillRepeatedly(Return(expected_file_size))
        ;

        EXPECT_CALL(*mock_request, GetFileName())
        .WillRepeatedly(Return(expected_file_name))
        ;

        EXPECT_CALL(*mock_request, GetMetaData())
        .WillRepeatedly(ReturnRef(expected_metadata))
        ;
    }

    EXPECT_CALL(*mock_request, GetStream())
    .WillRepeatedly(Return(expected_stream))
    ;

    EXPECT_CALL(*mock_request, GetDataID())
    .WillRepeatedly(Return(expected_id))
    ;

    EXPECT_CALL(*mock_request, GetOpCode())
    .WillRepeatedly(Return(op_code))
    ;

    if (op_code == asapo::Opcode::kOpcodeTransferDatasetData) {
        EXPECT_CALL(*mock_request, GetCustomData_t())
        .WillRepeatedly(Return(expected_custom_data))
        ;
    }
}

MessageMeta DbCheckRequestHandlerTests::PrepareMessageMeta() {
    MessageMeta message_meta;
    message_meta.size = expected_file_size;
    message_meta.name = expected_file_name;
    message_meta.id = expected_id;
    message_meta.buf_id = expected_buf_id;
    message_meta.source = expected_host_uri;
    message_meta.metadata = expected_metadata;
    message_meta.stream = expected_stream;
    message_meta.timestamp = std::chrono::system_clock::now();
    return message_meta;
}


TEST_F(DbCheckRequestHandlerTests, ErrorIfRecordsDoNotMatch) {
    expected_message_meta.metadata = "{\"meta\":2}";
    ExpectRequestParams(asapo::Opcode::kOpcodeTransferData, expected_data_source);

    MockGetBrokerUri();
    MockGetData(expected_message_meta.Json());

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(asapo::ReceiverErrorTemplates::kBadRequest));
}

TEST_F(DbCheckRequestHandlerTests, DuplicateErrorIfRecordsMatch) {
    ExpectRequestParams(asapo::Opcode::kOpcodeTransferData, expected_data_source);

    MockGetBrokerUri();
    MockGetData(expected_message_meta.Json());

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(ReceiverErrorTemplates::kWarningDuplicatedRequest));
}

TEST_F(DbCheckRequestHandlerTests, DuplicateErrorIfRecordsMatchWithEmptyMetadata) {
    expected_message_meta.metadata = "{}";
    expected_metadata = "{}";
    ExpectRequestParams(asapo::Opcode::kOpcodeTransferData, expected_data_source, true);

    MockGetBrokerUri();
    MockGetData(expected_message_meta.Json());

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(ReceiverErrorTemplates::kWarningDuplicatedRequest));

}

TEST_F(DbCheckRequestHandlerTests, OkIfNotFound) {
    ExpectRequestParams(asapo::Opcode::kOpcodeTransferData, expected_data_source);

    MockGetBrokerUri();
    MockGetData(expected_message_meta.Json(), HttpCode::Conflict);

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(nullptr));
}

TEST_F(DbCheckRequestHandlerTests, ErrorIfDbError) {
    ExpectRequestParams(asapo::Opcode::kOpcodeTransferData, expected_data_source);

    MockGetBrokerUri();
    MockGetData(expected_message_meta.Json(), HttpCode::ServiceUnavailable);

    auto err = handler.ProcessRequest(mock_request.get());
    ASSERT_THAT(err, Eq(ReceiverErrorTemplates::kInternalServerError));
}

}
