#include <gtest/gtest.h>
#include <gmock/gmock.h>

#include "asapo/unittests/MockIO.h"
#include "asapo/unittests/MockDatabase.h"
#include "asapo/unittests/MockLogger.h"
#include "asapo/unittests/MockHttpClient.h"

#include "../../src/receiver_error.h"
#include "../../src/request.h"
#include "../../src/request_handler/request_factory.h"
#include "../../src/request_handler/request_handler.h"
#include "../../src/request_handler/request_handler_db_meta_write.h"

#include "../mock_receiver_config.h"
#include "asapo/common/data_structs.h"
#include "asapo/common/networking.h"
#include "asapo/common/internal/version.h"
#include "../receiver_mocking.h"

using namespace testing;
using namespace asapo;

namespace {

class DbMetaWriterHandlerTests : public Test {
  public:
    std::string expected_collection_name = asapo::kDBMetaCollectionName;
    RequestHandlerDbMetaWrite handler{};
    std::unique_ptr<NiceMock<MockRequest>> mock_request;
    NiceMock<asapo::MockLogger> mock_logger;
    ReceiverConfig config;
    std::string expected_data_source = "source";
    std::string expected_beamtime_id = "beamtime";
    std::string expected_stream = "stream";
    std::string meta_str =
        R"("info":"stat","int_data":0,"float_data":0.1,"bool":false)";
    const uint8_t* expected_meta = reinterpret_cast<const uint8_t*>(meta_str.c_str());
    uint64_t expected_meta_size = meta_str.size();
    std::string expected_meta_name = "bt";
    CustomRequestData expected_custom_data{0, 0, 0};

    std::string expected_token = "token";
    std::string expected_server_uri = "localhost:8400";
    std::string expected_broker_uri = "asapo-broker:5005";
    std::string expected_consumer_protocol = asapo::kConsumerProtocol.GetVersion();
    std::string expected_broker_protocol = asapo::kConsumerProtocol.GetBrokerVersion();
    std::string expected_broker_api = expected_broker_uri + "/" + expected_broker_protocol;
    std::string expected_stream_encoded = "str%20%24%20eam%24";

    std::string expected_pipeline_step_id = "pipeline_step_id";
    std::string expected_producer_instance_id = "producer_instance_id";

    NiceMock<MockHttpClient> mock_http_client{};
    void SetUp() override {
        GenericRequestHeader request_header;
        request_header.data_id = 0;
        handler.service_request_->httpclient__ = std::unique_ptr<asapo::HttpClient> {&mock_http_client};
        handler.log__ = &mock_logger;
        mock_request.reset(new NiceMock<MockRequest> {request_header, 1, "", nullptr, nullptr});
        ON_CALL(*mock_request, GetPipelineStepId()).WillByDefault(ReturnRef(expected_pipeline_step_id));
        ON_CALL(*mock_request, GetProducerInstanceId()).WillByDefault(ReturnRef(expected_producer_instance_id));
        ON_CALL(*mock_request, GetBeamtimeId()).WillByDefault(ReturnRef(expected_beamtime_id));
        ON_CALL(*mock_request, GetDataSource()).WillByDefault(ReturnRef(expected_data_source));
        ON_CALL(*mock_request, GetStream()).WillByDefault(Return(expected_stream));
        ON_CALL(mock_http_client, UrlEscape_t(expected_stream)).WillByDefault(Return(expected_stream_encoded));
        ON_CALL(mock_http_client, UrlEscape_t("default")).WillByDefault(Return("default_encoded"));
        // ON_CALL(*mock_request, GetFileName()).WillByDefault(Return(expected_file_name));

        SetDefaultRequestCalls(mock_request.get(),expected_beamtime_id);
    }
    void TearDown() override {
        handler.service_request_.release();
        Mock::VerifyAndClear(&mock_http_client);
    }

    void ExpectRequestParams(const std::string& ver, uint64_t mode, const std::string& stream);
    void MockGetServiceUri(std::string service, std::string result) {
        EXPECT_CALL(mock_http_client, Get_t(HasSubstr(expected_server_uri + "/asapo-discovery/v0.1/" + service + "?token="
                                                      + expected_token + "&protocol=" + expected_consumer_protocol), _, _)
                                            ).WillOnce(DoAll(
                                                    SetArgPointee<1>(HttpCode::OK),
                                                    SetArgPointee<2>(nullptr),
                                                    Return(result)));
    }

    void MockGetBrokerUri() {
        MockGetServiceUri("asapo-broker", expected_broker_uri);
    }
};

MATCHER_P(M_CheckIngestMode, mode, "") {
    return arg.op ==  mode.op && arg.upsert == mode.upsert;
}


void DbMetaWriterHandlerTests::ExpectRequestParams(const std::string& ver, uint64_t mode, const std::string& stream) {
    EXPECT_CALL(*mock_request, GetDataSource())
    .WillRepeatedly(ReturnRef(expected_data_source))
    ;

    EXPECT_CALL(*mock_request, GetDataSize())
    .WillRepeatedly(Return(expected_meta_size))
    ;

    EXPECT_CALL(*mock_request, GetData())
    .WillRepeatedly(Return((void*)expected_meta))
    ;

    EXPECT_CALL(*mock_request, GetApiVersion())
    .WillRepeatedly(Return(ver))
    ;

    if (mode > 0) {
        EXPECT_CALL(*mock_request, GetStream())
        .WillRepeatedly(Return(stream))
        ;
        expected_custom_data[asapo::kPosMetaIngestMode] = mode;
        EXPECT_CALL(*mock_request, GetCustomData_t()).WillRepeatedly(Return(expected_custom_data));
    }
}

TEST_F(DbMetaWriterHandlerTests, CallsIngestBeamtimeMeta) {
    SetReceiverConfig(config, "none");
    ExpectRequestParams(asapo::GetReceiverApiVersion(), 11, "");
    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client, Post_t(expected_broker_api + "/beamtime/beamtime//default_encoded/write_meta"
                                         + "?token=token&instanceid=&pipelinestep=&id=0&mode_op=1&mode_upsert=true&document_id=bt", _,
                                         _, _, _)).WillOnce(DoAll(
                                                     SetArgPointee<3>(HttpCode::OK),
                                                     SetArgPointee<4>(nullptr),
                                                     Return("")
                                                 ));

    handler.ProcessRequest(mock_request.get());
}

TEST_F(DbMetaWriterHandlerTests, CallsIngestStreamMeta) {
    SetReceiverConfig(config, "none");

    ExpectRequestParams(asapo::GetReceiverApiVersion(), 13, expected_stream);

    MockGetBrokerUri();
    EXPECT_CALL(mock_http_client, Post_t(expected_broker_api + "/beamtime/beamtime//"+expected_stream_encoded+"/write_meta"
                                         + "?token=token&instanceid=&pipelinestep=&id=0&mode_op=3&mode_upsert=true&document_id=st_"+
                                         expected_stream, _,
                                         _, _, _)).WillOnce(DoAll(
                                                     SetArgPointee<3>(HttpCode::OK),
                                                     SetArgPointee<4>(nullptr),
                                                     Return("")
                                                 ));

    handler.ProcessRequest(mock_request.get());
}



}
