#include <gtest/gtest.h>
#include <gmock/gmock.h>
#include <thread>

#include "../src/data_cache.h"


using namespace testing;

using asapo::DataCache;
using asapo::CacheMeta;


namespace {


class DataCacheTests : public Test {
  public:
    uint64_t expected_cache_size = 1024 * 1024;
    uint64_t expected_size = 10;
    uint64_t expected_val = 1;
    float expected_keepunlocked_ratio = 0.2f;
    CacheMeta* meta1;
    CacheMeta* meta2;
    DataCache cache{expected_cache_size, expected_keepunlocked_ratio};
    void SetUp() override {
    }
    void TearDown() override {
    }
};

TEST_F(DataCacheTests, GetFreeSlotFailsDueToSize) {
    asapo::Error err;
    auto addr = cache.GetFreeSlotAndLock(expected_cache_size + 1, &meta1, "b", "so", "st", &err);
    ASSERT_THAT(addr, Eq(nullptr));
    ASSERT_THAT(err, Ne(nullptr));
}

void set_array(uint8_t* addr, uint64_t size, uint8_t val) {
    for (uint64_t i = 0; i < size; i++) {
        addr[i] = val;
    }
}

TEST_F(DataCacheTests, GetFreeSlotOK) {
    asapo::Error err;
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(1, &meta1, "b", "so", "st", &err);
    set_array(addr, 1, 2);
    ASSERT_THAT(addr[0], Eq(2));
    ASSERT_THAT(meta1->id, Gt(0));
}

TEST_F(DataCacheTests, GetFreeSlotStartsFromLastPointer) {
    asapo::Error err;
    uint8_t* ini_addr = (uint8_t*) cache.GetFreeSlotAndLock(1, &meta1, "b", "so", "st", &err);
    set_array(ini_addr, 1, 2);
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(1, &meta2, "b", "so", "st", &err);
    set_array(addr, 1, 1);
    ASSERT_THAT(ini_addr[0], Eq(2));
    ASSERT_THAT(ini_addr[1], Eq(1));
    ASSERT_THAT(meta1->id, Ne(meta2->id));
}

TEST_F(DataCacheTests, GetFreeSlotLocks) {
    asapo::Error err;
    uint8_t* ini_addr = (uint8_t*) cache.GetFreeSlotAndLock(1, &meta1, "b", "so", "st", &err);
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);
    ASSERT_THAT(ini_addr, Ne(nullptr));
    ASSERT_THAT(addr, Eq(nullptr));
    ASSERT_THAT(meta1, Ne(nullptr));
    ASSERT_THAT(meta2, Eq(nullptr));
}

TEST_F(DataCacheTests, GetFreeSlotStartsFromBeginIfNotFit) {
    asapo::Error err;
    uint8_t* ini_addr = (uint8_t*) cache.GetFreeSlotAndLock(1, &meta1, "b", "so", "st", &err);
    auto id = meta1->id;
    set_array(ini_addr, 1, 2);
    cache.UnlockSlot(meta1);
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);
    set_array(addr, expected_cache_size, 1);
    ASSERT_THAT(ini_addr[0], Eq(1));
    ASSERT_THAT(id, Ne(meta2->id));
}

TEST_F(DataCacheTests, GetFreeSlotCannotWriteIfAlreadyWriting) {
    asapo::Error err;
    cache.GetFreeSlotAndLock(1, &meta1, "b", "so", "st", &err);
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);
    ASSERT_THAT(addr, Eq(nullptr));
    ASSERT_THAT(meta2, Eq(nullptr));

}

TEST_F(DataCacheTests, PrepareToReadIdNotFound) {
    uint64_t id;
    id = 0;
    uint8_t* addr = (uint8_t*) cache.GetSlotToReadAndLock(id, expected_size, &meta1);
    ASSERT_THAT(addr, Eq(nullptr));
    ASSERT_THAT(meta1, Eq(nullptr));
}

TEST_F(DataCacheTests, PrepareToReadOk) {
    asapo::Error err;
    auto data_size = static_cast<uint64_t>(static_cast<double>(expected_cache_size) * 0.7);
    uint8_t* ini_addr = (uint8_t*) cache.GetFreeSlotAndLock(data_size, &meta1, "b1", "so1", "st1", &err);

    uint8_t* addr = (uint8_t*) cache.GetSlotToReadAndLock(meta1->id, data_size, &meta2);
    ASSERT_THAT(addr, Eq(ini_addr));
    ASSERT_THAT(meta1, Eq(meta2));
    ASSERT_THAT(meta2->size, Eq(data_size));

    ASSERT_THAT(meta1->beamtime, Eq("b1"));
    ASSERT_THAT(meta1->source, Eq("so1"));
    ASSERT_THAT(meta1->stream, Eq("st1"));

    ASSERT_THAT(meta2->beamtime, Eq("b1"));
    ASSERT_THAT(meta2->source, Eq("so1"));
    ASSERT_THAT(meta2->stream, Eq("st1"));
}


TEST_F(DataCacheTests, PrepareToReadFailsIfTooCloseToCurrentPointer) {
    asapo::Error err;
    auto data_size = static_cast<uint64_t>(static_cast<double>(expected_cache_size) * 0.9);
    cache.GetFreeSlotAndLock(data_size, &meta1, "b", "so", "st", &err);

    uint8_t* addr = (uint8_t*) cache.GetSlotToReadAndLock(meta1->id, data_size, &meta2);
    ASSERT_THAT(addr, Eq(nullptr));

}

TEST_F(DataCacheTests, GetFreeSlotRemovesOldMetadataRecords) {
    DataCache cache{expected_cache_size, 0};
    asapo::Error err;
    CacheMeta* meta3, *meta4, *meta5;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(10, &meta1, "b1", "so1", "st1", &err);
    cache.GetFreeSlotAndLock(10, &meta2, "b2", "so2", "st2", &err);
    cache.GetFreeSlotAndLock(expected_cache_size - 30, &meta3, "b3", "so3", "st3", &err);
    auto id1 = meta1->id;
    auto id2 = meta2->id;
    auto id3 = meta3->id;

    cache.UnlockSlot(meta1);
    cache.UnlockSlot(meta2);
    cache.UnlockSlot(meta3);

    cache.GetFreeSlotAndLock(10, &meta4, "b4", "so4", "st4", &err);
    auto id4 = meta4->id;

    cache.UnlockSlot(meta4);
    cache.GetFreeSlotAndLock(30, &meta5, "b5", "so5", "st5", &err);
    uint8_t* addr1 = (uint8_t*) cache.GetSlotToReadAndLock(id1, 10, &meta);
    uint8_t* addr2 = (uint8_t*) cache.GetSlotToReadAndLock(id2, 10, &meta);
    uint8_t* addr3 = (uint8_t*) cache.GetSlotToReadAndLock(id3, expected_cache_size - 30, &meta);
    uint8_t* addr4 = (uint8_t*) cache.GetSlotToReadAndLock(id4, 10, &meta);

    ASSERT_THAT(addr1, Eq(nullptr));
    ASSERT_THAT(addr2, Eq(nullptr));
    ASSERT_THAT(addr3, Eq(nullptr));
    ASSERT_THAT(addr4, Ne(nullptr));
    ASSERT_THAT(meta->size, Eq(10));
    ASSERT_THAT(meta->beamtime, Eq("b4"));
    ASSERT_THAT(meta->source, Eq("so4"));
    ASSERT_THAT(meta->stream, Eq("st4"));
}

TEST_F(DataCacheTests, GetFreeSlotRemovesOldWhenCrossTheBoundary) {
    asapo::Error err;
    DataCache cache{expected_cache_size, 0};
    CacheMeta* meta1, *meta2, *meta3;
    CacheMeta* meta4, *meta5, *meta;
    auto addr1_alloc = cache.GetFreeSlotAndLock(expected_cache_size / 3 - 1, &meta1, "b", "so", "st", &err);
    auto addr2_alloc = cache.GetFreeSlotAndLock(expected_cache_size / 3 - 1, &meta2, "b", "so", "st", &err);
    auto addr3_alloc = cache.GetFreeSlotAndLock(expected_cache_size / 3 - 1, &meta3, "b", "so", "st", &err);
    auto id1 = meta1->id;
    auto id2 = meta2->id;
    auto id3 = meta3->id;
    cache.UnlockSlot(meta1);
    cache.UnlockSlot(meta2);
    cache.UnlockSlot(meta3);
    auto addr4_alloc = cache.GetFreeSlotAndLock(expected_cache_size / 2 + 5, &meta4, "b", "so", "st", &err);
    auto id4 = meta4->id;
    cache.UnlockSlot(meta4);
    auto addr5_alloc = cache.GetFreeSlotAndLock(expected_cache_size / 2 + 5, &meta5, "b", "so", "st", &err);
    auto id5 = meta5->id;

    uint8_t* addr1 = (uint8_t*) cache.GetSlotToReadAndLock(id1, expected_cache_size / 3 - 1, &meta);
    uint8_t* addr2 = (uint8_t*) cache.GetSlotToReadAndLock(id2, expected_cache_size / 3 - 1, &meta);
    uint8_t* addr3 = (uint8_t*) cache.GetSlotToReadAndLock(id3, expected_cache_size / 3 - 1, &meta);
    uint8_t* addr4 = (uint8_t*) cache.GetSlotToReadAndLock(id4, expected_cache_size / 2 + 5, &meta);
    uint8_t* addr5 = (uint8_t*) cache.GetSlotToReadAndLock(id5, expected_cache_size / 2 + 5, &meta);

    ASSERT_THAT(addr1_alloc, Ne(nullptr));
    ASSERT_THAT(addr2_alloc, Ne(nullptr));
    ASSERT_THAT(addr3_alloc, Ne(nullptr));
    ASSERT_THAT(addr4_alloc, Ne(nullptr));
    ASSERT_THAT(addr5_alloc, Ne(nullptr));

    ASSERT_THAT(addr1, Eq(nullptr));
    ASSERT_THAT(addr2, Eq(nullptr));
    ASSERT_THAT(addr3, Eq(nullptr));
    ASSERT_THAT(addr4, Eq(nullptr));
    ASSERT_THAT(addr5, Ne(nullptr));


    ASSERT_THAT(meta->size, Eq(expected_cache_size / 2 + 5));
}



TEST_F(DataCacheTests, GetSlotToReadSizeOk) {
    asapo::Error err;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(expected_size, &meta1, "b", "so", "st", &err);

    uint8_t* addr1 = (uint8_t*) cache.GetSlotToReadAndLock(meta1->id, expected_size, &meta);

    ASSERT_THAT(addr1, Ne(nullptr));
    ASSERT_THAT(meta->size, Eq(expected_size));
    ASSERT_THAT(meta->beamtime, Eq("b"));
    ASSERT_THAT(meta->source, Eq("so"));
    ASSERT_THAT(meta->stream, Eq("st"));
}

TEST_F(DataCacheTests, GetSlotToReadWrongSize) {
    asapo::Error err;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(expected_size, &meta1, "b", "so", "st", &err);

    uint8_t* addr1 = (uint8_t*) cache.GetSlotToReadAndLock(meta1->id, expected_size + 1, &meta);

    ASSERT_THAT(addr1, Eq(nullptr));
}



TEST_F(DataCacheTests, CannotGetFreeSlotIfNeedCleanOnebeingReaded) {
    CacheMeta* meta;
    asapo::Error err;

    uint8_t* ini_addr = (uint8_t*) cache.GetFreeSlotAndLock(10, &meta1, "b", "so", "st", &err);
    auto res = cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    uint8_t* addr = (uint8_t*) cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);

    ASSERT_THAT(ini_addr, Ne(nullptr));
    ASSERT_THAT(res, Eq(ini_addr));
    ASSERT_THAT(addr, Eq(nullptr));
}


TEST_F(DataCacheTests, CanGetFreeSlotIfWasUnlocked) {
    asapo::Error err;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(10, &meta1, "b", "so", "st", &err);
    cache.UnlockSlot(meta1);
    cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    cache.UnlockSlot(meta);
    auto addr = cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);

    ASSERT_THAT(addr, Ne(nullptr));
}

TEST_F(DataCacheTests, IncreasLockForEveryRead) {
    asapo::Error err;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(10, &meta1, "b", "so", "st", &err);
    cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    cache.UnlockSlot(meta);
    auto addr = cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);

    ASSERT_THAT(addr, Eq(nullptr));
}

TEST_F(DataCacheTests, DecreasLockForEveryUnlock) {
    asapo::Error err;
    CacheMeta* meta;
    cache.GetFreeSlotAndLock(10, &meta1, "b", "so", "st", &err);
    cache.UnlockSlot(meta1);

    cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    cache.GetSlotToReadAndLock(meta1->id, 10, &meta);
    cache.UnlockSlot(meta);
    cache.UnlockSlot(meta);
    auto addr = cache.GetFreeSlotAndLock(expected_cache_size, &meta2, "b", "so", "st", &err);

    ASSERT_THAT(addr, Ne(nullptr));
}


TEST_F(DataCacheTests, GetFreeSlotCreatesCorrectIds) {
    asapo::Error err;
    CacheMeta* meta3, *meta4;
    cache.GetFreeSlotAndLock(10, &meta1, "b", "so", "st", &err);
    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    cache.GetFreeSlotAndLock(10, &meta2, "b", "so", "st", &err);
    std::this_thread::sleep_for(std::chrono::milliseconds(10));
    cache.GetFreeSlotAndLock(10, &meta3, "b", "so", "st", &err);
    std::this_thread::sleep_for(std::chrono::milliseconds(1));
    cache.GetFreeSlotAndLock(10, &meta4, "b", "so", "st", &err);

    auto c1 = static_cast<uint32_t>(meta1->id);
    auto c2 = static_cast<uint32_t>(meta2->id);
    auto c3 = static_cast<uint32_t>(meta3->id);
    auto c4 = static_cast<uint32_t>(meta4->id);

    auto t1 = meta1->id >> 32;
    auto t2 = meta2->id >> 32;
    auto t3 = meta3->id >> 32;
    auto t4 = meta4->id >> 32;

    ASSERT_THAT(c2, Eq(c1 + 1));
    ASSERT_THAT(c3, Eq(c2 + 1));
    ASSERT_THAT(c4, Eq(c3 + 1));

    ASSERT_THAT(t2, Ne(t1));
    ASSERT_THAT(t2, Ne(t3));
    ASSERT_THAT(t4, Ne(t3));
    ASSERT_THAT(t1, Ne(t3));
    ASSERT_THAT(t4, Ne(t2));
    ASSERT_THAT(t4, Ne(t1));

}

}
