import pytest
import json
import numpy as np
import os
from time import time
import logging
import sys
from datetime import datetime
from pathlib import Path


log = logging.getLogger(__name__)

# Using the environment variable `BUILD_DIR` enables to run the tests from a directory
# different from the directory where `ASAP::O` was built. In particular, the GitLab CI
# job `test-services-linux-debug` defined in `.gitlab-ci.yml` runs the tests from the
# project root directory.
if build_dir_env := os.environ.get("BUILD_DIR"):
    build_dir = Path(build_dir_env)
    assert build_dir.exists()
else:
    build_dir = Path()

sys.path.insert(0, str(build_dir / "producer/api/python/"))
sys.path.insert(0, str(build_dir / "consumer/api/python/"))

import asapo_producer
import asapo_consumer


def callback(payload, err):
    if err is not None and not isinstance(err, asapo_producer.AsapoServerWarning):
        # the data was not sent. Something is terribly wrong.
        log.error(f"could not send: {err}")
        raise Exception('Sending data failed')
    elif err is not None:
        # The data was sent, but there was some unexpected problem, e.g. the file was overwritten.
        log.debug(f"sent with warning: {err}")
    else:
        # all fine
        log.debug(f"successfuly sent: {payload.get('id')}")


@pytest.fixture(autouse=True)
def cleanup_db(asapo_config):
    log.info(f"Run test for consumer {asapo_consumer.__version__} and producer {asapo_producer.__version__}")
    os.makedirs(asapo_config.data_path, exist_ok=True)
    yield None
    os.system(f"rm -fr {asapo_config.data_path}")


@pytest.fixture
def producer(asapo_config):
    return asapo_producer.create_producer(asapo_config.host, 'processed', asapo_config.beamtime, "auto",
                                          asapo_config.data_source,
                                          asapo_config.token, 1, 6000)


@pytest.fixture
def consumer(asapo_config):
    return asapo_consumer.create_consumer(asapo_config.host, "auto",
                                          False, asapo_config.beamtime, asapo_config.data_source,
                                          asapo_config.token, 100)


def send_data(producer, idx, stream, substream=None, n_modules=3):
    producer.send(idx,  # message number. Should be unique and ordered.
                  f"processed/tilt_4_008_{stream}_{substream}_data_00000{idx}.h5",
                  np.ones((1000, 1000)).astype(np.uint8).tobytes(),
                  callback=callback,
                  user_meta=json.dumps({"ggg": 123}),
                  dataset=None if substream is None else (substream, n_modules),
                  # ingest_mode=asapo_producer.CACHE_ONLY_INGEST_MODE,
                  stream=stream
                  )


def test_get_stream_meta(producer, consumer):
    stream = f"{time()}"

    producer.send_stream_meta(json.dumps({'aaa': 123, 'bbb': 567}),
                              mode='update',
                              callback=callback,
                              stream=stream)
    producer.wait_requests_finished(2000)

    meta = producer.get_stream_meta(stream=stream)
    log.info(f"Stream meta {meta}")
    assert meta['aaa'] == 123
    assert meta['bbb'] == 567

    producer.send_stream_meta(json.dumps({'ccc': 234}),
                              mode='update',
                              callback=callback,
                              stream=stream)
    producer.wait_requests_finished(2000)

    meta = producer.get_stream_meta(stream=stream)
    log.info(f"Stream meta {meta}")
    assert meta['aaa'] == 123
    assert meta['bbb'] == 567
    assert meta['ccc'] == 234

    producer.send_stream_meta(json.dumps({'abc': 123}),
                              mode='replace',
                              callback=callback,
                              stream=stream)
    producer.wait_requests_finished(2000)

    meta = producer.get_stream_meta(stream=stream)
    log.info(f"Stream meta {meta}")
    assert meta['abc'] == 123
    assert meta.get('bbb') is None


def test_get_bt_meta(producer, consumer):
    md = json.dumps({'dataset_substream': 123})
    producer.send_beamtime_meta(md,
                                mode='update',
                                callback=callback)
    producer.wait_requests_finished(50000)
    meta = consumer.get_beamtime_meta()
    log.info(f"Beamtime meta {meta}")


def test_delete_stream_producer(consumer, producer):
    stream = f"{time()}_2"
    for i_msg in [5, 4, 3, 2, 1]:
        send_data(producer, i_msg, stream)
    producer.wait_requests_finished(50000)

    g = consumer.generate_group_id()
    _, message = consumer.get_next(g, meta_only=True,
                                   stream=stream, ordered=False)

    producer.delete_stream(stream)

    with pytest.raises(asapo_consumer.AsapoEndOfStreamError):
        _, message = consumer.get_next(g, meta_only=True,
                                       stream=stream, ordered=False)


def test_get_stream_info(consumer, producer):
    inset = f'test_{time()}'
    str_list = [f'{inset}_{5 - i}' for i in range(5)]
    for stream in str_list:
        send_data(producer, 1, stream)
        producer.wait_requests_finished(50000)

    stream_info = producer.stream_info(str_list[0])
    log.info(f"Stream info: {stream_info}")
    assert stream_info['name'] == str_list[0]

    stream_info = producer.last_stream()
    log.info(f"Last stream: {stream_info}")
    assert stream_info['name'] == str_list[4]
